# Keysight_357670A driver, TW 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import qkit
from qkit.core.instrument_base import Instrument
import time
import numpy as np
from qkit.drivers.visa_prologix import instrument


class Keysight_35670A(instrument, Instrument):
    """
    This is the python driver for the Keysight 35670A signal analyzer
    Usage:
    Initialise with
    <name> = instruments.create('<name>', address='<GPIB address>', reset=<bool>)
    """
    def __init__(self, name, address, reset=False, **kwargs):
        """
        Initializes 

        Input:
            name (string)    : name of the instrument
            address (string) : GPIB address
        """

        print('Initialization will take some seconds... Do not worry')
        self.ip = kwargs.get("ip", "10.22.197.155")
        super(Keysight_35670A, self).__init__(address, timeout=10, chunk_size=int(4096), delay=1e-3, ip=self.ip)
        Instrument.__init__(self, name, tags=['physical'])

        self.data_type = {'power': 'XFR:POW', 'linear': 'XFR:POW:LIN',
                          'coherence': 'XFR:POW:COH', 'cross': 'XFR:POW:CROS',
                          'time': 'XTIM:VOLT', 'histogram': 'XTIM:VOLT:HIST'}
        self.inv_data_type = {v: k for k, v in self.data_type.items()}

        # Implement parameters
        # ToDO: insert all parameters and function from the script below, check how to deal with channels
        self.add_parameter('instrument_mode', type=str, flags=Instrument.FLAG_GETSET)
        self.add_parameter('nop', type=int, flags=Instrument.FLAG_GET,minval=101, maxval=801)
        self.add_parameter('averages', type=int, flags=Instrument.FLAG_GETSET, minval=1, maxval=99999)
        self.add_parameter('Average', type=bool, flags=Instrument.FLAG_GETSET)
        self.add_parameter('centerfreq', type=float, flags=Instrument.FLAG_GETSET,minval=0, maxval=51.2e3, units='Hz', tags=['sweep'])
        self.add_parameter('startfreq', type=float, flags=Instrument.FLAG_GETSET, minval=0, maxval=51.2e3, units='Hz', tags=['sweep'])
        self.add_parameter('stopfreq', type=float, flags=Instrument.FLAG_GETSET, minval=0, maxval=51.2e3, units='Hz', tags=['sweep'])
        self.add_parameter('span', type=float, flags=Instrument.FLAG_GETSET, minval=0, maxval=51.2e3, units='Hz')
        self.add_parameter('frequency_resolution', type=int, flags=Instrument.FLAG_GETSET, minval=100, maxval=800)
        self.add_parameter('sweeptime', type=float, flags=Instrument.FLAG_GETSET, minval=0, maxval=10000, units='s')
        self.add_parameter('sweeptime_averages', type=float, flags=Instrument.FLAG_GET, minval=0, maxval=1e-3, units='s')
        self.add_parameter('window_type', type=str, flags=Instrument.FLAG_GETSET)
        self.add_parameter('active_traces', type=int, flags=Instrument.FLAG_GETSET)
        self.add_parameter('repeat_averaging', type=bool, flags=Instrument.FLAG_GETSET)
        
        self.add_function('get_freqpoints')
        self.add_function('get_tracedata')
        self.add_function('get_single_trace')
        self.add_function('pre_measurement')
        self.add_function('start_measurement')
        self.add_function('ready')
        self.add_function('post_measurement')
        self.add_function('calibration')
        self.add_function('abort_measurement')
        self.add_function('set_measurement_continous')
        self.add_function('measure_PSD')
        self.add_function('fast_averaging')
        self.add_function('set_display_format')
        self.add_function('set_ac_coupling')
        
        if reset is True:
            self._set_GPIB_dev_reset()
            time.sleep(5)

        self.frequency_changed = True
        self.device_type='signal_analyzer'
        self.instrument_mode = None
        self.set_instrument_mode('FFT')
        self.set_return_format()
        qkit.flow.sleep(1)
        self.get_all()

    def get_all(self):
        if self.instrument_mode == 'FFT':
            self.get_centerfreq()
            self.get_span()
            self.get_frequency_resolution()
            self.get_startfreq()
            self.get_stopfreq()
            self.get_window_type()
        self.get_instrument_mode()
        self.get_averages()
        self.get_sweeptime()
        self.get_sweeptime_averages()
        self.get_active_traces()
        self.get_nop()
        self.y_unit_c1=self.get_y_unit(1)
        self.y_unit_c2=self.get_y_unit(2)

    def ready(self):
        """This is a proxy function, returning True when the signal analyzer has finished measuring / averaging."""
        if (self._read_operation_status_register() & 256) == 0:
            return True
        else:
            return False

    def set_return_format(self, s_digits=6):
        '''specifies the number of significant digits'''
        self.write('FORMAT:DATA ASCii, {}'.format(s_digits))
    
    def calibration(self):
        reply=self.ask('*cal?')
        if reply is not 0:
            print("calibration error!")

    def auto_calibration(self, status):
        self.write('calibration:auto {}'.format(status))

    def abort_measurement(self):
        self.write('abor')

    def start_measurement(self):
        """starts a new measurement"""
        self.write("INIT:IMM")
        self.frequency_changed = False

    def set_measurement_continous(self, pause):
        """pauses(=0) or resumes(=1) the measurement"""
        self.write("INIT:CONT {}".format(pause))

    def _read_operation_status_register(self):
        """reads the operation status register, returns??"""
        return int(self.ask("STAT:OPER:COND?")[1:-1])

    def pre_measurement(self):
        self.get_all()

    def post_measurement(self):
        """nothing to do here??"""
        pass

    def do_set_active_traces(self, num_of_traces):
        """
        sets the number of active traces
        """
        trace_names='ABCD'
        self.write("CALC:ACT "+trace_names[:num_of_traces])

    def do_get_active_traces(self):
        """gets the active traces"""
        return len(self.ask('CALC:ACT?')[:-1])

    def get_trace_name(self, trace):
        """
        function for signal spectroscopy script to automatically generate names for trace data
        :param trace:
        :return: (string) name of trace
        """
        meas_typ= self.get_meas_data_type(trace)
        meas_format = self.get_trace_format(trace)
        return meas_typ + '_' + meas_format

    def set_input_channels(self, channel, status):
        """
        Sets the active input channels. Especially useful if you only have one channel and want to go to the highest
        possible frequency, i.e., 102.4kHz
        :param channel: 1 or 2 (int)
        :param status: True or False (bool)
        :return: None
        """
        if status == 0 and channel == 1 :
            print("channel 1 must always be on")
            return
        state = ['OFF', 'ON']
        self.write(':inp{} {}'.format(channel, state[status]))

    def set_ac_coupling(self, channels=[1,2], val=[True, True]):
        """
        sets input coupling to AC or DC for specified channels. 3dB point is at 1 Hz for AC coupling
        :param channels: channels to change (list)
        :param val: True for AC coupling False for DC coupling (List)
        :return:
        """
        mode = ['DC', 'AC']
        for num, i in enumerate(channels):
            self.write(':Input{}:Coup {}'.format(i, mode[val[num]]))

    def get_ac_coupling(self, channel):
        """
        gets input coupling either AC or DC for specified channels. 3dB point is at 1 Hz for AC coupling
        :param channel: channels to ask
        :return:
        """
        return self.ask(':Input{}:Coup?'.format(channel))


    #### Frequency options ####

    def do_set_centerfreq(self, center_freq): # in Hz
        """sets the center frequency"""
        self.write("Freq:Cent {}".format(center_freq))
        self.frequency_changed = True

    def do_get_centerfreq(self):
        """gets the center frequency"""
        #print self.ask("Freq:Cent?")
        self.centerfreq=float(self.ask("Freq:Cent?")[:-1])
        return self.centerfreq

    def do_set_startfreq(self, start_freq):
        """sets the start frequency"""
        self.write("Freq:Start {}".format(start_freq))
        self.frequency_changed = True

    def do_get_startfreq(self):
        """gets the start frequency"""
        self.startfreq=float(self.ask("Freq:Start?")[:-1])
        return self.startfreq

    def do_set_span(self, span):
        """sets the frequency span"""
        self.write("Freq:Span {}".format(span))
        self.frequency_changed = True

    def do_get_span(self):
        """gets the frequency span"""
        self.frequency_span=float(self.ask("Freq:Span?")[:-1])
        return self.frequency_span

    def do_set_frequency_resolution(self, resolution):
        """sets the frequency resolution in lines (unitless)"""
        if resolution not in [100, 200, 400, 800]:
            print('only 100, 200, 400, 800 are allowed values')
        else:
            self.write("Freq:Res {}".format(resolution))
            self._nop=resolution+1

    def do_get_frequency_resolution(self):
        """gets the frequency resolution"""
        return int(float(self.ask("Freq:Res?")[:-1]))

    def do_get_nop(self):
        """proxy function for spectroscopy script"""
        if self.instrument_mode == 'FFT':
            self._nop= self.get_frequency_resolution()+1
        elif self.instrument_mode == 'histogram':
            pass  # maybe sth must go in here
        return self._nop

    def do_set_stopfreq(self, stop):
        """sets the stop frequency"""
        self.write("Freq:Stop {}".format(stop))
        self.frequency_changed = True

    def do_get_stopfreq(self):
        """gets the stop frequeny"""
        self.stopfreq=float(self.ask("Freq:Stop?")[:-1])
        return self.stopfreq

    def set_frequency_bins(self):
        '''sets frequency bin in histogram mode, not yet written'''
        pass

    def do_set_instrument_mode(self, mode):
        """sets the instrument mode, expects string: 'FFT', 'histogram', 'correlation' """
        if mode == 'FFT':
            mode_number = 0
        elif mode == 'histogram':
            mode_number = 4
        elif mode == 'correlation':
            mode_number = 5
        else:
            print('wrong input')
            return
        if self.do_get_instrument_mode() != mode:
            self.frequency_changed = True  # new x-Axis when instrument mode is changed
        self.instrument_mode = mode
        self.write('INST:NSEL {}'.format(mode_number))

    def do_get_instrument_mode(self):
        """returns the instrument mode"""
        mode_number = int(self.ask('INST:NSEL?')[1:-1])
        if mode_number == 0:
            self.instrument_mode = 'FFT'
        elif mode_number == 4:
            self.instrument_mode = 'histogram'
        return self.instrument_mode

    def do_set_sweep_mode(self, mode):
        """sets sweepmode to mode='AUTO' or 'MAN' """
        if mode not in ['AUTO', 'MAN']:
            print('wrong input only AUTO and MAN are valid')
        else:
            self.write('swe:mode '+mode)

    def do_set_sweeptime(self, time):
        """sets the sweep time (recording time) in sec  sweeptime=freqspan/freqresolution.
        Chaning either will change the other"""
        self.write('swe:TIME {}'.format(time))

    def do_get_sweeptime(self):
        """gets the sweep time (recording time) """
        self.sweeptime=float(self.ask('swe:TIME?')[:-1])
        return self.sweeptime

    def do_get_sweeptime_averages(self):
        self.sweeptime_averages = self.do_get_sweeptime()*self.do_get_averages()
        return self.sweeptime_averages

    def do_set_window_type(self, type):
        """sets the window type, expects string either 'HANN', 'FLAT' or 'UNIF'
            HANN smears least but might be unprecise in amplitude, the opposite
            is the case for FLAT
        """
        if type not in ['HANN', 'FLAT', 'UNIF']:
            print("wrong input only HANN, FLAT, UNIF are allowed")
        else:
            self.write('WINDOW '+type)

    def do_get_window_type(self):
        """gets the window type"""
        self.window_type = self.ask('WINDOW?')[:-1]
        return self.window_type

    def set_autoscale(self, trace, option):
        """turns autoscalle off and on. A trace has to be selected,
        option must be a string [AUTO, OFF, ON, ONCE] or 0 and 1"""
        self.write('disp:wind{}trac:y:auto '.format(trace)+option)

    def set_input_shielding(self, mode, channel = [1,2]):
        """ determines the input shielding mode (expects string), either "ground" or "floating",
        for the chosen channels"""
        for i in channel:
            self.write('INP{}:LOW '.format(i)+mode)

    def get_input_shielding(self, channel):
        """ returns the input shield of the chosen channel"""
        return self.ask('INP{}:LOW?'.format(channel))[:-1]

    def set_range(self, value, channel=[1,2]):
        """sets a fixed autorange, expects channel as list and an input value in Volts peak"""
        for i in channel:
            self.write('SENS:VOLT{}:DC:RANGE {} VPK'.format(i, value))

    def get_range(self, channel=[1,2]):
        """returns the specified range (Vpk) as a list for the chosen channels"""
        reply = []
        for i in channel:
            reply.append(self.ask('SENS:VOLT{}:DC:RANGE?'.format(i)))
        return reply

    def set_autorange(self, mode, channel=[1,2]):
        """turns (1) on or off (0) the range of the chosen channel to auto"""
        for i in channel:
            self.write('sens:volt{}:dc:range:auto {}'.format(i, int(mode)))

    def get_autorange(self, channels=[1,2]):
        """gets the autorange for specified channel; 1 is on 0 is off"""
        reply = []
        for i in channels:
            reply.append(bool(self.ask('sens:volt{}:dc:range:auto?'.format(i))[1]))  # remove newline and + before by[1]
        return reply

    def set_display_format(self, number):
        """selects the number of traces [1,2,4] to be displayed (3 is not possible)"""
        option={1: 'SINGLE', 2: 'ULOWER', 4:'QUAD'}
        self.write('DISPLAY:FORM ' + option[number])

    def set_trace_format(self, trace, option):
        """selects the display format for the specified trace, i.e., how it is displayed and to be collected from the
        signal analyzer options: linear, phase, real, imag, NYQuist, UPHase, GDElay, POLAR. Don't confuse linear here
        with the chosen power spectrum in meas datatype. It means that it's not logarithmic"""
        self.write('calc{}:format {}'.format(trace, option))

    def get_trace_format(self, trace):
        """
        gets the trace format of the selected trace, ie. the format how the data is diplayed and will be collected
        from the signal analyzer
        :param trace: 1 to 4
        :return: trace format (string)
        """
        return self.ask('calc{}:format?'.format(trace))[0:-1].lower()

    def set_meas_data_type(self, option, channel, trace):
        """
        sets the datatype that is measured. You can choose between power spectrum, linear spectrum (interesting
        if you want the complex values [no power averaging here]), cross spectrum (i.e. FFT(covariance ch1/ch2)),
        time record. coherence (you have to look that up yourself). If you want to have the unfiltered timetrace,
        you have to set the instrument into histogram mode first.
        :param option: (str) 'power', 'linear, 'coherence', 'cross', 'time', 'histogram'
        :param channel: input channel (1 or 2), for cross spectrum and coherence, both channels are used (input '1,2')
        :param trace: trace to be displayed (1 to 4)
        :return: None
        """
        if option in ['coherence', 'cross']:
            channel = '1,2'  # This is channel 1 and 2. Device expects a string
        if trace > 2:
            self.set_display_format(4)
        message= "CALC{}:FEED '{} {}'".format(trace, self.data_type[option], channel)
        self.write(message)

    def get_meas_data_type(self, trace):
        """
        :param trace: trace to ask for meas_data_type (1 to 4)
        :return: the type of measurement data i.e. lin spec or power spec
        """
        reply = self.ask("CALC{}:FEED?".format(trace))
        meas_data = self.inv_data_type[reply.split()[0][1:]]+'_ch_'+reply.split()[1][:-1]  # removing the ' in the reply
        return meas_data

    ########## AVERAGING ##############

    def do_set_Average(self, status):
        """turns averagin off/on"""
        if status not in [0,1]:
            print("wrong input 0=off, 1=on")
        else:
            self.set_measurement_continous(0)  # pauses measurement so that new sequence does not start automatically
            self.write('AVER {}'.format(int(status)))

    def do_get_Average(self):
        return bool(int(self.ask('AVER?')[1:-1]))

    def fast_averaging(self, status, displaynumber=None):
        """turns on fast averaging, i.e., display is only updated at the chosen displaynumber"""
        if status not in [0,1]:
            print("wrong input 0=off, 1=on")
            return
        if status == 1:
            self.write('AVERAGE:IRES:RATE {}'.format(displaynumber))
            self.write('Average:IRES:Stat {}'.format(status))
        else:
            self.write('Average:Ires:Stat {}'.format(status))

    def do_set_averages(self, number):
        """sets the number of averages"""
        self.write('AVER:COUN {}'.format(number))

    def do_get_averages(self):
        """gets the numbers of averages"""
        if self.instrument_mode == 'histogram':
            self.averages = 1
        else:
            self.averages = int(float(self.ask('AVER:COUN?')[:-1]))
        return self.averages

    def do_set_repeat_averaging(self, status):
        """specify what happens after average_number is reached, i.e., repeat measurements or wait"""
        if status == 0:
            self.write('AVERAGE:TCON FREEZE')
        elif status == 1:
            self.write('AVERAGE:TCON REPEAT')
        else:
            print("wrong input 0=HOLD, 1=REPEAT")

    def do_get_repeat_averaging(self):
        repeat_aver=self.ask('AVER:TCON?')[:-1]
        if repeat_aver == 'FRE':
            return False
        if repeat_aver[:3] == 'REP':
            return True

   ########## MATH-OPTIONS #########

    def math_expression(self, register, expression):
        """stores a math expression (string conform with the device options) into the chosen register"""
        self.write('Calculate:Math:Expr{} '.format(register) + expression)

    def math_state(self, channel, status):
        """activates math evalution for chosen schannel"""
        self.write('calculate{}:math:state {}'.format(channel, status))

    def math_select(self, channel, register):
        """selects a previously defined math_expression in a register for a channel"""
        self.write(':calc{}:math:select F{}'.format(channel, register))

    def calculate_PSD(self, channel):
        """Calculates and display the PSD for the chosen channel"""
        expression='(PSD(PSPEC{}))'.format(channel)
        self.math_expression(channel, expression)
        self.math_state(channel, 1)
        self.math_select(channel, channel)

    ######## Getting the data ######

    def get_freqpoints(self):
        """returns the x_values of a trace as numpy-array"""
        if self.instrument_mode == 'FFT':
            if self.frequency_changed:
                self.set_averages(1)
                self.do_set_repeat_averaging(0)
                self.start_measurement()  # device only sets new x-axis when new measurement has ended
                qkit.flow.sleep(0.05)
                while not self.ready():  # but spectroscopy asks for x-values before meausurment is started
                    qkit.flow.sleep(0.1)
            self.x_values=np.fromstring(self.ask("CALC:X:DATA?"), dtype=np.float32, sep=',')
            self.set_averages(self.averages)
            return self.x_values[0:self.get_nop()]
        elif self.instrument_mode == 'histogram':
            if self.frequency_changed:
                self.start_measurement()  # device only sets new x-axis when new measurement has ended
                qkit.flow.sleep(0.05)
                while not self.ready():  # but spectroscopy asks for x-values before meausurment is started
                    qkit.flow.sleep(0.1)
            self.x_values = np.fromstring(self.ask("CALC:X:DATA?"), dtype=np.float32, sep=',')
            return self.x_values
        else:
            self.x_values = np.fromstring(self.ask("CALC:X:DATA?"), dtype=np.float32, sep=',')
            return self.x_values

    def get_single_trace(self, trace):
        """returns the trace of trace 1,2,3,4"""
        trace_values=np.fromstring(self.ask("CALC{}:DATA?".format(trace)), dtype=np.float32, sep=',')
        return trace_values

    def get_tracedata(self, trace, type='AmpPha'):
        return self.get_single_trace(trace)

    def get_tracedata_pseudo_vna_scan(self):
        """
        This is a function to perform a pseudo vna scan with the spectrums analyzer using the timetrace (histogram)
        mode.
        :return: amplitude, phase, I and Q data (one single point)
        """
        I = self.get_single_trace(1)
        Q = self.get_single_trace(2)
        I_max = (self.x_values[np.argmax(I)])
        Q_max = (self.x_values[np.argmax(Q)])
        amp = np.absolute(I_max + 1j * Q_max)
        pha = np.angle(I_max + 1j * Q_max)
        return [amp], [pha], [I_max], [Q_max]

    def set_y_unit(self, unit, channel):
        """sets the y-unit (string), can also be used for PSD calucalation"""
        if unit[0] != '"':
            unit = '"' + unit + '"'
        if channel in [1, 2, 3, 4]:  # implement string for unit
            self.write('calculate{}:unit:voltage '.format(channel) + unit)

    def get_y_unit(self, trace):
        """gets the y-unit of the selected trace"""
        if trace in [1,2,3,4]:
            y_unit=self.ask('CALCulate{}:UNIT:VOLTage?'.format(trace))[:-1]
            return y_unit.strip('"')
        else:
            print('wrong input')

    def measure_PSD(self, channel):
        """proxy funciton for set y-unit with unit = "V2/Hz" """
        self.set_y_unit('"V2/HZ"', channel)



#!/usr/bin/env python

# Substantially derived from code by (c) 2015 Richard Hull The MIT License (MIT)
# https://github.com/rm-hull/ssd1306:
#       "Permission is hereby granted, free of charge, to any person obtaining a copy
#       "of this software and associated documentation files (the "Software"), to deal
#       "in the Software without restriction, including without limitation the rights
#       "to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#       "copies of the Software, and to permit persons to whom the Software is
#       "furnished to do so, subject to the following conditions:
#       "The above copyright notice and this permission notice shall be included in
#       "all copies or substantial portions of the Software."
#
# B Lavery 2015:
# This derivative "library" module is not installed to the python system as Hull's version was:
# it simply resides alongside your own python script.
# In this version, the I2C bus object needs to be handed in as a parameter from your user code.
# This makes this one same library file work with either Raspberry Pi or Virtual GPIO system.
# Hull's (clever) auto-displaying "canvas" is replaced by a persistent draw object
# which can be incrementally changed. This canvas needs coded "display()" calls to push to the hardware.

from PIL import Image, ImageDraw


class ssd1306():

    def __init__(self, bus, address=0x3C):
        self.cmd_mode = 0x00
        self.data_mode = 0x40
        self.bus = bus
        self.addr = address
        self.width = 128
        self.height = 64
        self.pages = int(self.height / 8)
        self.image = Image.new('1', (self.width, self.height))
        self.canvas = ImageDraw.Draw(self.image) # this is a "draw" object for preparing display contents

        self._command(
            const.DISPLAYOFF,
            const.SETDISPLAYCLOCKDIV, 0x80,
            const.SETMULTIPLEX,       0x3F,
            const.SETDISPLAYOFFSET,   0x00,
            const.SETSTARTLINE,
            const.CHARGEPUMP,         0x14,
            const.MEMORYMODE,         0x00,
            const.SEGREMAP,
            const.COMSCANDEC,
            const.SETCOMPINS,         0x12,
            const.SETCONTRAST,        0xCF,
            const.SETPRECHARGE,       0xF1,
            const.SETVCOMDETECT,      0x40,
            const.DISPLAYALLON_RESUME,
            const.NORMALDISPLAY,
            const.DISPLAYON)

    def _command(self, *cmd):
        """
        Sends a command or sequence of commands through to the
        device - maximum allowed is 32 bytes in one go.
        LIMIT ON ARDUINO: CMD BYTE + 31 = 32, SO LIMIT TO 31     bl
        """
        assert(len(cmd) <= 31)
        self.bus.write_i2c_block_data(self.addr, self.cmd_mode, list(cmd))

    def _data(self, data):
        """
        Sends a data byte or sequence of data bytes through to the
        device - maximum allowed in one transaction is 32 bytes, so if
        data is larger than this it is sent in chunks.
        In our library, only data operation used is 128x64 long, ie whole canvas.
        """

        for i in range(0, len(data), 31):
            self.bus.write_i2c_block_data(self.addr, self.data_mode, list(data[i:i+31]))


    def display(self):
        """
        The image on the "canvas" is flushed through to the hardware display.
        Takes the 1-bit image and dumps it to the SSD1306 OLED display.
        """

        self._command(
            const.COLUMNADDR, 0x00, self.width-1,  # Column start/end address
            const.PAGEADDR,   0x00, self.pages-1)  # Page start/end address

        pix = list(self.image.getdata())
        step = self.width * 8
        buf = []
        for y in range(0, self.pages * step, step):
            i = y + self.width-1
            while i >= y:
                byte = 0
                for n in range(0, step, self.width):
                    byte |= (pix[i + n] & 0x01) << 8
                    byte >>= 1

                buf.append(byte)
                i -= 1

        self._data(buf) # push out the whole lot

    def cls(self):
        self.canvas.rectangle((0, 0, self.width-1, self.height-1), outline=0, fill=0)
        self.display()

    def onoff(self, onoff):
        if onoff == 0:
            self._command(const.DISPLAYOFF)
        else:
            self._command(const.DISPLAYON)


class const:
    CHARGEPUMP = 0x8D
    COLUMNADDR = 0x21
    COMSCANDEC = 0xC8
    COMSCANINC = 0xC0
    DISPLAYALLON = 0xA5
    DISPLAYALLON_RESUME = 0xA4
    DISPLAYOFF = 0xAE
    DISPLAYON = 0xAF
    EXTERNALVCC = 0x1
    INVERTDISPLAY = 0xA7
    MEMORYMODE = 0x20
    NORMALDISPLAY = 0xA6
    PAGEADDR = 0x22
    SEGREMAP = 0xA0
    SETCOMPINS = 0xDA
    SETCONTRAST = 0x81
    SETDISPLAYCLOCKDIV = 0xD5
    SETDISPLAYOFFSET = 0xD3
    SETHIGHCOLUMN = 0x10
    SETLOWCOLUMN = 0x00
    SETMULTIPLEX = 0xA8
    SETPRECHARGE = 0xD9
    SETSEGMENTREMAP = 0xA1
    SETSTARTLINE = 0x40
    SETVCOMDETECT = 0xDB
    SWITCHCAPVCC = 0x2



import numpy as np
from zhinst.toolkit.control.drivers import UHFQA as _UHFQA
from zhinst.toolkit.control.node_tree import Node, Parameter
from qkit.core.instrument_base import Instrument
from qkit.drivers.ZHInst_Abstract import ZHInst_Abstract


class ZHInst_UHFQA(ZHInst_Abstract):
    def __init__(self, name, serialnumber, host="129.13.93.38",  **kwargs):
        super().__init__(name, **kwargs)
        self._uhfqa = _UHFQA(name, serialnumber, host=host)
        self._uhfqa.setup()
        self._uhfqa.connect_device()
        
        # Iterate node tree for readable entries and hook them into QKit
        self.blacklist = ["awg_sequencer", "awg_waveform", "awg_elf", "awg_dio", "elf", "qa_result_statistics",
                          "scope_wave", "auxin_sample", "dio_input", "system_fwlog", "features_code"]
        self.mount_api("node_dump_uhfqa.txt", self._uhfqa.nodetree)

        # Register readout methods
        self.add_function("get_qubit_result", channels=(0, 9))
        self.add_function("enable_channel", channels=(0, 9))
        self.add_function("disable_channel", channels=(0, 9))
        self.add_function("arm")
        self.add_function("compile_program")
        self.add_function("run")
        self.add_parameter("readout_frequency", flags=Instrument.FLAG_GETSET, channels=(0, 9), type=float)
        self.add_parameter("readout_amplitude", flags=Instrument.FLAG_GETSET, channels=(0, 9), type=float)

    def compile_program(self, *args, **kwargs):
        self._uhfqa.awg.set_sequence_params(*args, **kwargs)
        self._uhfqa.awg.compile()

    def get_qubit_result(self, channel=None):
        if channel == None: # TODO: Register this correctly in QKit?
            return [self._uhfqa.channels[i].result() for i in range(10)]
        return self._uhfqa.channels[channel].result()

    def enable_channel(self, channel):
        self._uhfqa.channels[channel].enable()

    def disable_channel(self, channel):
        self._uhfqa.channels[channel].disable()

    def arm(self, *args, **kwargs):
        self._uhfqa.arm(*args, **kwargs)

    def run(self):
        self._uhfqa.awg.run()

    def _do_set_readout_frequency(self, frequency, channel):
        self._uhfqa.channels[channel].readout_frequency(frequency)

    def _do_set_readout_amplitude(self, amplitude, channel):
        self._uhfqa.channels[channel].readout_frequency(amplitude)

    def _do_get_readout_frequency(self, channel):
        return self._uhfqa.channels[channel].readout_frequency()

    def _do_get_readout_amplitude(self, channel):
        return self._uhfqa.channels[channel].readout_frequency()
    
        



#!/usr/bin/env python
from __future__ import unicode_literals

#Copyright (c) 2017 Justin Winokur
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.
# HR: imported into qkit 28.1.2018 from 
# https://github.com/Jwink3101/list_dict_DB
#
# (pip install list_dict_DB is also possible)


import copy
from collections import defaultdict
import time
import types

class list_dict_DB(object):

    def __init__(self,items=None,attributes=None,default_attribute=None,  \
                    exclude_attributes=None,                              \
                    allowMultipleEdit=False,alwaysReturnList=True,        \
                    indexObjects=False):
        """
        Create a multi-item DB from a list of dictionaries that may be
        queried by any specified attribute.

        This is useful since, once created, lookup/query/"in" checks are
        O(1), Creation is still O(N)

        Note: Unless an entry is changed with update(), it must be reindexed

        Inputs:
        --------

        items  [ *empty* ] (list)
            List of dictionaries with each attribute
        
        attributes [None] (list, None)
            Either a list of attributes to index, or if specified as None
            (default) it will add every attribute of every item and assign 
            missing ones to be `default_attribute` (unless excluded)
        
            NOTE: If attributes is set, you may still add items with extra
                  attributes. They just won't be indexed. 
                  Or use add_attribute()
        
        exclude_attributes [ *empty* ] (list)
            Attributes that shouldn't ever be added even if attributes=None 
            for dynamic addition of attributes.
        
        default_attribute [None] (*any*)
            Default attribute to assign if `attributes=None` and it is missing.
            If the specified object is callable, it will call it (for example,
            `default_attribute=list` will make it an empty list), Otherwise,
            it will set it to whatever value is specified.
        
        Options: (These may be changed later too)
        --------
        
        allowMultipleEdit [False] (bool)
            Allows the remove() method to delete multiple items and and 
            update() method to update multiple items
        
        alwaysReturnList [True] (bool)
            If set to True, will always return a list
        
        indexObjects: [False]
            If True, will automatically take any object and use its
            __dict__ as the dict. 
            
            Note 
                * Changing to False after adding an object will cause issues.
                * Does not support __slots__ since they are immutable
            

        Additional Opperations:
        ----------------------

        This supports index-lookup with a dictionary as well as
        a python `in` check and lookup by a dictionary

        Example
        -------
            see list_dict_DB.print_example()

        Tips:
        ------

            * You can simply dump the DB with JSON using the DB.items()
              and then reload it with a new DB
            
            * There is also an attribute called `_index` which can be used to
              query by index.

        """
        
        # Handle inputs
        if items is None:
            items = list()

        if exclude_attributes is None:
            exclude_attributes = list()

        self.allowMultipleEdit = allowMultipleEdit
        self.alwaysReturnList = alwaysReturnList        
        self.indexObjects = indexObjects


        self.attributes = attributes # Will be reset in first add        
        self._is_attr_None = attributes is None
        self.default_attribute = default_attribute
        self.exclude_attributes = exclude_attributes

        self.N = 0 # Will keep track
        self._list = []

        self._empty = _emptyList()
        self._ix = set()

        # Add the items
        for item in items:
            self.add(item)
        
        self._i = 0 # Counter for iterator if not called with iteritems
        
        self._time = time.time()
    
        # Edge case: No items
        if self.attributes is None:
            self.attributes = []
        
    def add(self,item):
        """
        Add an item or items to the DB
        """
        if isinstance(item,(list,tuple,types.GeneratorType)):
            for it in item:
                self.add(it)
            return
        
        # handle other object types
        item0 = item
        item = self._convert2dict(item)
        
        if self.N == 0:
            attributes = self.attributes
            if attributes is None:
                attributes = list(item.keys())

            self.attributes = [attrib for attrib in attributes \
                                if attrib not in self.exclude_attributes] # Make a copy
            

            # Set up the lookup
            self._lookup = {attribute:defaultdict(list) for attribute in self.attributes}
        
        ix = len(self._list) # The length will be 1+ the last ix so do not change this

        if self._is_attr_None: # Set to None which means we add all
            for attrib in item.keys():
                if attrib in self.exclude_attributes:
                    continue
                if attrib not in self.attributes:
                    self.add_attribute(attrib,self.default_attribute)
        # Add built in ones
        for attrib in self.attributes:
            if attrib not in item:
                if hasattr(self.default_attribute, '__call__'):
                    item[attrib] = self.default_attribute()
                else:
                    item[attrib] = self.default_attribute
                    
            value = item[attrib]
            self._append(attrib,value,ix)

        # Finally add it
        self._list.append(item0)
        self.N += 1
        self._ix.add(ix)

    def query(self,*A,**K):
        """
        Query the value for attribute. Will always return a
        single item if only one matches, otherwise it will return a list.

        Tip: If you ALWAYS want a list, do:
            list(DB.query(...))
        
        Usage
        -----     
        
        Any combination of the following will works
        
        Keywords: Only check equality
                   
        >>> DB.query(attrib=val)               
        >>> DB.query(attrib1=val1,attrib2=val2)  # Match both
        
        >>> DB.query({'attrib':val})                  
        >>> DB.query({'attrib1':val1,'attrib2':val2}) # Match Both
                                      
        Query Objects (DB.Q, DB.Qobj)
        
        >>> Q = DB.Q() # Load Qobj from list_dict_DB
        >>> DB.query(DQ.attrib == val)            
        >>> DB.query( (Q.attrib1 == val1) &  (Q.attrib1 == val2) )  # Parentheses are important!
        >>> DB.query( (Q.attrib1 == val1) &  (Q.attrib1 != val2))
                                   
        """
        ixs = self._ixs(*A,**K)

        if len(ixs) == 1 and not self.alwaysReturnList:
            return self._list[ixs[0]]
        else:
            return [self._list[ix] for ix in ixs]
        

    def isin(self,*A,**K):
        """
        Check if there is at least one item that matches the given query
        
        see query() for usage
        """

        return len(self._ixs(*A,**K))>0

    def reindex(self,*args):
        """
        Reindex the dictionary for specified attributes (or all)
        
        Usage
        -----
        
        >>> DB.reindex()                # All
        >>> DB.reindex('attrib')        # Reindex 'attrib'
        >>> DB.reindex('attrib1','attrib2') # Multiple
        
        See Also
        --------
            update() method which does not require reindexing
        """
        if len(args) == 0:
            attributes = self.attributes
            
            # Just an extra check (and makes a copy)
            attributes = [attr for attr in attributes \
                        if attr not in self.exclude_attributes]
        else:
            attributes = args
            if any(a in self.exclude_attributes for a in args):
                raise ValueError('Cannot reindex an excluded attribute') 

        for attribute in attributes:
            self._lookup[attribute] = defaultdict(list) # Reset
        
        for ix,item in enumerate(self._list):
            if item is None: continue
            item = self._convert2dict(item)
            for attrib in attributes:
                value = item[attrib]
                self._append(attrib,value,ix)
    
    def update(self,*args,**queryKWs):
        """
        Update an entry without needing to reindex the DB (or a specific 
        attribute)
        
        Usage:
        ------
        
        >>> DB.update(updated_dict, query_dict_or_Qobj, query_attrib1=val1,...)
        >>> DB.update(updated_dict, query_attrib1=val1,...)
        
        Inputs:
        -------
        
        updated_dict : Dictionary with which to update the entry. This is
                       done using the typical dict().update() construct to
                       overwrite it
        
        query_dict_or_Qobj   
                     : Either the dictionary used in the query or a Qobj that 
                       defines a more advanced query
        
        query_attrib1=val1
                     : Additional (or sole) query attributes
    
        DB Options:
        -----------
        This is set at instantiation but can be changed 'DB' is the list_dict_DB

        DB.allowMultipleEdit
                        - If True, will update multiple items
    
        Notes:
        ------
            * Updating an item requires a deletion in a list that has length n 
              equal to the number of items matching an attribute. This is O(n).
              However changing the entry directly and reindexing is O(N) where
              N is the size of the DB. If many items are changing and you do not
              need to query them in between, it *may* be faster to directly
              update the item and reindex
        """
        
        if len(args) == 1:
            updated_dict = args[0]
            query = {}
        elif len(args) == 2:
            updated_dict,query = args
        else:
            raise ValueError('Incorrect number of inputs. See documentation')
        
        updated_dict = self._convert2dict(updated_dict)
        if not isinstance(updated_dict,dict):
            raise ValueError('Must specify updated values as a dictionary')
        
        query = self._convert2dict(query)
        if isinstance(query,Qobj):
            ixs = self._ixs(query,**queryKWs)
        elif isinstance(query,dict):
            queryKWs.update(query)
            ixs = self._ixs(**queryKWs)
        else:
            raise ValueError('Unrecognized query {:s}. Must be a dict or Qobj',format(type(query)))
        
        if len(ixs) == 0:
            raise ValueError('Query did not match any results')
        
        if len(ixs)>1 and not self.allowMultipleEdit:
            raise ValueError("Query returned multiple results. Set 'allowMultipleEdit' or change query")
        
        for ix in ixs:
            # Get original item
            item = self._list[ix]
            item = self._convert2dict(item)
            
            # Allow the update to also include non DB attributes.
            # The intersection will eliminate any exclude_attributes
            attributes = set(updated_dict.keys()).intersection(self.attributes)
            
            for attrib in attributes: # Only loop over the updated attribs
                # get old value
                value = item[attrib]
                
                # Remove any ix matching it
                self._remove(attrib,value,ix)
                
                # Get new value
                value = updated_dict[attrib]
                
                # Add ix to any new value
                self._append(attrib,value,ix)
                
            # Update the item
            item.update(updated_dict)
        
        return
        
        
        
    def add_attribute(self,attribute,*default):
        """
        Add an attribute to the index attributes.

        Usage
        -----
        >>> DB.add_attribute('new_attrib') # Will raise an error if *any*
                                           # items don't have 'new_attrib'
        >>> DB.add_attribute('new_attrib',default)
                                           # Set any missing to the default
        
        If the `default` is callable, it will call it instead. (such as `list`
        to add an empty list)
        
        """
        if attribute in self.exclude_attributes:
            raise ValueError("Can't add exclude_attributes")
        
        attrib = attribute
        if not hasattr(self,'_lookup'):
            self._lookup = {}
        self._lookup[attribute] = defaultdict(list)

        set_default = False
        if len(default) >0:
            set_default = True
            default = default[0]

        for ix,item in enumerate(self._list):
            if item is None: continue
            item = self._convert2dict(item)
            try:
                value = item[attribute]
                self._append(attrib,value,ix)
            except KeyError as KE:
                if set_default:
                    if hasattr(default, '__call__'):
                        item[attribute] = default()
                    else:
                        item[attribute] = default
                else:
                    raise KeyError("Attribute {:s} not found".format(attrib))

                value = item[attribute]
                self._append(attrib,value,ix)

        self.attributes.append(attribute)

    def remove(self,*A,**K):
        """
        Remove item that matches a given attribute or dict. See query() for
        input specification


        DB Options:
        This is set at instantiation but can be changed directly
        -----------
        
        DB.allowMultipleEdit
                        - If True, this can remove multiple matching items

        """

        ixs = self._ixs(*A,**K)

        if len(ixs)>1 and not self.allowMultipleEdit:
            raise ValueError("Query returned multiple results. Set 'allowMultipleEdit' or change query")

        if len(ixs) == 0:
            raise ValueError('No matching items')

        items = []

        for ix in list(ixs): # Must remove it from everything.
            # not sure what is happening, but it seems that I need to make a copy
            # since Python is doing something strange here...

            item = self._list[ix]
            item = self._convert2dict(item)

            for attrib in self.attributes:
                value = item[attrib]
                self._remove(attrib,value,ix)
                
            # Remove it from the list by setting to None. Do not reshuffle
            # the indices. A None check will be performed elsewhere
            self._list[ix] = None
            self._ix.difference_update([ix])
            self.N -= 1
    
    def items(self):
        """
        Return a list of items.
        """
        return [a for a in self._list if a is not None]

    def iteritems(self):
        """
        Return an iterator over the items.
        """
        return (a for a in self._list if a is not None)
    
    def Qobj(self):
        """
        Query object already loaded with the DB
        
            DB.Qobj() <==> DB.Qobj() <==> Qobj(DB)
        """
        return Qobj(self) 
    
    def Q(self):
        """
        same as Qobj method
        """
        return Qobj(self)
    
    def _convert2dict(self,obj):
        """ 
        Convert objects to a regular dictionary for the sake of indexing
        
        Also return Qobjs untouched since they may be used in queries too
        
        If it not an Qobj or dict, it will try to get it's __dict__ and if 
        that doesn't work, will just return it
        """
        if isinstance(obj,Qobj):
            return obj
        
        if isinstance(obj,dict): # Also accounts for OrderedDicts or ...
            return obj           # ... anything that inherits dict
        
        if self.indexObjects and hasattr(obj,'__dict__'):
            return obj.__dict__
            
        return obj
        
    
    def _ixs(self,*args,**kwords):
        """
        Get the inde(x/ies) of matching information
        """
        if not hasattr(self,'_lookup') or self.N==0: # It may be empty
            return []
 
        # Make the entire kwords be lists with default of []. Edge case of
        # multiple items
        for key,val in kwords.items():
            if not isinstance(val,list):
                kwords[key] = [val]
        kwords = defaultdict(list,kwords)
        
        Q = Qobj(self) # Empty object
        for arg in args:
            arg = self._convert2dict(arg) # handle other object types
            if isinstance(arg,Qobj):
                Q = Q & arg # Will add these conditions. If Q is empty, will just be arg
                continue
            if isinstance(arg,dict):
                for key,val in arg.items(): # Add it rather than update in case it is already specified
                    kwords[key].append(val)
            else:
                raise ValueError('unrecognized input of type {:s}'.format(str(type(arg))))
        
        # Construct a query for kwords
        for key,value in kwords.items():
            if isinstance(value,list) and len(value) == 0:
                value = [self._empty]
            for val in _makelist(value):
                Qtmp = Qobj(self)
                Qtmp._attr = key
                Q = Q & (Qtmp == val)

        ixs = Q._ixs
        # Ensure one match
        if ixs is None: 
            ixs = []
        return list(ixs)        
        
    
    def _index(self,ix):
        """
        Return ix if it hasn't been deleted
        """
        try:
            item = self._list[ix]
        except IndexError:
            return []
        
        if item is None:
            return []
        
        return [ix]
    
    def _append(self,attrib,value,ix):
        """
        Add to the lookup and update the modify time
        """
        # Final check but we should be guarded from this
        if attrib in self.exclude_attributes:
            print('BAD! Should guard against this in public methods!')
            raise ValueError('Cannot reindex an excluded attribute') 
        
        valueL = _makelist(value)
        for val in valueL: 
            self._lookup[attrib][val].append(ix)
        if len(valueL) == 0:
            self._lookup[attrib][self._empty].append(ix) # empty list
        self._time = time.time()
    
    def _remove(self,attrib,value,ix):
        """
        Remove from the lookup and update the modify time
        """
        
        valueL = _makelist(value)
        for val in valueL: 
            self._lookup[attrib][val].remove(ix)
        if len(valueL) == 0:
            self._lookup[attrib][self._empty].remove(ix) # empty list
    
        self._time = time.time()
    
    def __contains__(self,check_diff):
        check_diff = self._convert2dict(check_diff)
        if not ( isinstance(check_diff,dict) or isinstance(check_diff,Qobj)):
            raise ValueError('Python `in` queries should be a of {attribute:value} or Qobj')
        return self.isin(check_diff)

    def __len__(self):
        return self.N

    def __getitem__(self,item):
        item = self._convert2dict(item)
        if isinstance(item,dict) or isinstance(item,Qobj):
            return self.query(item)
        elif isinstance(item,int): # numbered item
            if self._list[item] is not None:
                return self._list[item]
            else:
                raise ValueError("Index has been deleted")
        else:
            raise ValueError("Must specify DB[{'attribute':val}] or DB[index]'")
    __call__ = query
    
    def __iter__(self):
        return self

    def __next__(self):
        while self._i < len(self._list):
            self._i += 1 # Increment it but then search back
            if self._list[self._i-1] is not None:
                return self._list[self._i-1]
        raise StopIteration()
    next = __next__ # For compatability
    
    
def _makelist(input):
    if isinstance(input,list):
        return input
    return [input] 

class _emptyList(object):
    def __init__(self):
        pass
    def __hash__(self):
        return 9999999999999
    def __eq__(self,other):
        return isinstance(other,list) and len(other)==0 
        
class Qobj(object):
    """
    Query objects. This is a bit wonky. 
    
    Calling 
        * Q.attribute sets attribute
        * Q.attribute == val (or any other comparison) set the index of elements
        * Q1 & Q1 or other boolean perform set operations
        
    Useful Methods:
        _filter : (or just `filter` if not an attribute): Apply a filter
                  to the DB
    """
    def __init__(self,DB,ixs=None,attr=None):
        self._DB = DB
        self._ixs = ixs
        self._attr = attr
        
        self._time = time.time()
        
    
    def _valid(self):
        if self._time < self._DB._time:
            raise ValueError('This query object is out of date from the DB. Create a new one')
    
    def _filter(self,filter_func):
        """
        
        If 'filter' is NOT an attribute of the DB, this can be called 
        with 'filter' instead of '_filter'
        
        Apply a filter to the data that returns True if it matches and False 
        otherwise
        
        Note that filters are O(N)
        """
        self._valid() # Actually, these would still work but still check
        ixs = set()
        for ix,item in enumerate(self._DB._list): # loop all
            item = self._DB._convert2dict(item)
            if item is None:
                continue
            if filter_func(item):
                ixs.add(ix)
        self._ixs = ixs
        return self.copy()
         
            
    # Comparisons   
    def __eq__(self,value):
        self._valid()
        
        if self._DB.N == 0:
            self._ixs = set()
            return self.copy()
        
        first_set = True
        for val in _makelist(value): # Account for list inputs
            if self._attr == '_index':
                if first_set:
                    ixs = set(self._DB._index(val))
                    first_set = False
                else:
                    ixs = ixs.intersection(self._DB._index(val))
                continue
            if self._attr not in self._DB.attributes:
                raise KeyError("'{:s}' is not an attribute".format(self._attr))
                
            ixs_at = self._DB._lookup[self._attr][val]
            if first_set:
                ixs = set(ixs_at)
                first_set = False
            else:
                ixs = ixs.intersection(ixs_at)
        
        self._ixs = ixs
        return self.copy()
     
    def __ne__(self,value):
        self._ixs = self._DB._ix - (self == value)._ixs
        return self.copy()
    
    def __lt__(self,value):
        self._valid() # Actually, these would still work but still check
        ixs = set()
        for ix,item in enumerate(self._DB._list): # loop all
            item = self._DB._convert2dict(item) 
            if item is None:
                continue
            for ival in _makelist(item[self._attr]):
                if ival < value:
                    ixs.add(ix)
        self._ixs = ixs
        return self.copy()

    def __le__(self,value):
        self._valid() # Actually, these would still work but still check
        ixs = set()
        for ix,item in enumerate(self._DB._list): # loop all
            item = self._DB._convert2dict(item)
            if item is None:
                continue
            if item[self._attr] <= value:
                ixs.add(ix)
        self._ixs = ixs
        return self.copy()
        
    def __gt__(self,value):
        self._valid() # Actually, these would still work but still check
        ixs = set()
        for ix,item in enumerate(self._DB._list): # loop all
            item = self._DB._convert2dict(item)
            if item is None:
                continue
            if item[self._attr] > value:
                ixs.add(ix)
        self._ixs = ixs
        return self.copy()
    
    def __ge__(self,value):
        self._valid() # Actually, these would still work but still check
        ixs = set()
        for ix,item in enumerate(self._DB._list): # loop all
            item = self._DB._convert2dict(item)
            if item is None:
                continue
            if item[self._attr] >= value:
                ixs.add(ix)
        self._ixs = ixs
        return self.copy()
    
    # Logic
    def __and__(self,Q2):               
        if self._ixs == None:  # An empty object and another will just return other
            return Q2
        self._ixs.intersection_update(Q2._ixs)
        return self.copy()
    def __or__(self,Q2):               
        self._ixs.update(Q2._ixs)
        return self.copy()
    def __invert__(self):
        self._ixs = self._DB._ix - self._ixs
        return self.copy()
    
    def __getattr__(self,attr):
        if attr == 'filter' and 'filter' not in self._DB.attributes:
            return self._filter
        self._attr = attr
        return self.copy()
    
    def copy(self):
        new = Qobj(self._DB,ixs=self._ixs,attr=self._attr)
        # Reset the time
        new._time = self._time
        return new
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    



#-*- coding: utf-8 -*-
from subprocess import Popen, PIPE
import os
import numpy as np
import logging
import json

from numpy.core.multiarray import ndarray

logging.basicConfig(level=logging.INFO)

import qkit
from qkit.storage import store
from qkit.storage.hdf_constants import ds_types
from qkit.core.lib.misc import str3,concat
import sys

try:
    if qkit.module_available("matplotlib"):
        import matplotlib.pyplot as plt
        from matplotlib.figure import Figure
        from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
        plot_enable = True
except AttributeError:
    try:
        import matplotlib.pyplot as plt
        from matplotlib.figure import Figure
        from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
        plot_enable = True
    except ImportError:
        plot_enable = False

# this is for live-plots
def plot(h5_filepath, datasets=[], refresh = 2, live = True, echo = False):
    """
    Opens a .h5-file with qviewkit in a python subprocess.
    
    This is the outermost function to open a h5 file with qviewkit.
    To be independent from other processes i.e. running measurement the viewer
    is opend in a python subprocess. The function opens a shell and executes
    the qviewkit main routine with the given arguments.

    Args:
        h5_filepath: String, absolute filepath.
        datasets: Optional arrays of dataset urls to be opened at start of 
            qviewkit, default: []
        refresh: Optional integer refreshing time in seconds for live view, 
            default: 2
        live: Boolean, optional for constant checks for updates. 
            default: True
        echo: Boolean, optional. Echo settings for debugging, default: False
    Return:
        Popen call with the correct commend that opens a shell and executes
        the main qviewkit routine with a given filename. If given, some
        datasets open automaticall.
    """
    # the plot engine for live plots is set in the environement
    plot_viewer = qkit.cfg.get('plot_engine', None)
    
    # the final return call should look sth. like this:
    # python -m qkit.gui.qviewkit.main -f [h5_filepath] -ds amplitude,phase -rt 2 -live
    
    cmd = [sys.executable]
    cmd.append('-m')
    cmd.append(plot_viewer)
    
    cmd.append('-f')
    cmd.append(h5_filepath)
    
    if datasets:
        cmd.append('-ds')
        ds = ""
        for s in datasets:
            ds += s+','
        cmd.append(ds[:-1])
    cmd.append('-rt')
    cmd.append(str(refresh))
    if live:
        cmd.append('-live')
    
    if echo:
        print("Qviewkit open cmd: "+ str(cmd))
        os.putenv("HDF5_USE_FILE_LOCKING",'FALSE')
        P = Popen(cmd, shell=False, stdout=PIPE)
        print(P.stdout.read())
        return P
    else:
        os.putenv("HDF5_USE_FILE_LOCKING", 'FALSE')
        return Popen(cmd, shell=False, stdout=PIPE, text=True)


# this is for saving plots
def save_plots(h5_filepath, comment='', save_pdf=False):
    """
    Saves plots of all datasets with default settings.
    
    Args:
        h5_filepath: String, absolute filepath.
        comment: Optional comment for the plots to be added to the filenames.
            default : ''
        save_pdf: Optional boolean setting for the output file type.
            default: False
    """
    h5plot(h5_filepath, comment=comment, save_pdf=save_pdf)


class h5plot(object):
    """
    h5plot class plots and saves all dataset in the h5 file.

    We recursively walk through the entry tree of datasets, check for their
    ds_type and plot the data in a default setting with matplotlib. The plots 
    are then saved in an 'image' folder beside the h5 file. 
    """
    y_data = None  # type: ndarray

    def __init__(self,h5_filepath, comment='', save_pdf=False):
        """Inits h5plot with a h5_filepath (string, absolute path), optional 
        comment string, and optional save_pdf boolean.
        """
        if not plot_enable or not qkit.module_available("matplotlib"):
            logging.warning("matplotlib not installed. I can not save your measurement files as png. I will disable this function.")
            qkit.cfg['save_png'] = False
        if not qkit.cfg.get('save_png',True):
            return
        self.comment = comment
        self.save_pdf = save_pdf
        self.path = h5_filepath

        filepath = os.path.abspath(self.path)   #put filepath to platform standards
        self.filedir  = os.path.dirname(filepath)   #return directory component of the given pathname, here filepath

        self.image_dir = os.path.join(self.filedir,'images')
        try:
            os.mkdir(self.image_dir)
        except OSError:
            logging.warning('Error creating image directory.')
            pass

        # open the h5 file and get the hdf_lib object
        self.hf = store.Data(self.path)

        # check for datasets
        for i, pentry in enumerate(self.hf['/entry'].keys()):
            key='/entry/'+pentry
            for j, centry in enumerate(self.hf[key].keys()):
                try:
                    self.key='/entry/'+pentry+"/"+centry
                    self.ds = self.hf[self.key]
                    if self.ds.attrs.get('save_plot', True):
                        self.plt() # this is the plot function
                except Exception as e:
                    print("Exception in qkit/gui/plot/plot.py while plotting")
                    print(self.key)
                    print(e)
        #close hf file
        self.hf.close()
        print('Plots saved in ' + self.image_dir)

    def plt(self):
        """
        Creates the matplotlib figure, checks for some metadata and calls
        the plotting with respect to the ds_type of the dataset.
        
        Args:
            self: Object of the h5plot class.
        Returns:
            No return variable. The function operates on the given object.
        """
        logging.info(" -> plotting dataset: "+str(self.ds.attrs.get('name')))

        self.ds_type = self.ds.attrs.get('ds_type', '')
        self.x_ds_url = self.ds.attrs.get('x_ds_url','')
        self.y_ds_url = self.ds.attrs.get('y_ds_url','')
        self.z_ds_url = self.ds.attrs.get('z_ds_url','')

        self.fig = Figure(figsize=(20,10),tight_layout=True)
        self.ax = self.fig.gca()
        self.canvas = FigureCanvas(self.fig)

        self._unit_prefixes = {24: 'Y', 21: 'Z', 18: 'E', 15: 'P', 12: 'T', 9: 'G', 6: 'M', 3: 'k', 0: '', -3: 'm', -6: u'µ', -9: 'n', -12: 'p', -15: 'f', -18: 'a', -21: 'z', -24: 'y'}
        self.plot_styles = {0:'-', 1:'.-', 2:'.'}

        if self.ds_type == ds_types['coordinate']:
            #self.plt_coord()
            return
        elif self.ds_type == ds_types['vector']:
            self.plt_vector()
        elif self.ds_type == ds_types['matrix']:
            self.plt_matrix()
        elif self.ds_type == ds_types['box']:
            self.plt_box()
        elif self.ds_type == ds_types['txt']:
            #self.plt_txt()
            return
        elif self.ds_type == ds_types['view']:
            self.plt_view()
        else:
            return

        ## Some labeling depending on the ds_type. The label variables are set
        ## in the respective plt_xxx() fcts.
        self.ax.set_xlabel(self.x_label)
        self.ax.set_ylabel(self.y_label)
        self.ax.xaxis.label.set_fontsize(20)
        self.ax.yaxis.label.set_fontsize(20)
        self.ax.ticklabel_format(useOffset=False)
        for i in self.ax.get_xticklabels():
            i.set_fontsize(16)
        for i in self.ax.get_yticklabels():
            i.set_fontsize(16)

        save_name = str(os.path.basename(self.filedir))[0:6] + '_' + self.key.replace('/entry/','').replace('/','_')
        if self.comment:
            save_name = save_name+'_'+self.comment
        image_path = str(os.path.join(self.image_dir,save_name))

        if self.save_pdf:
            self.canvas.print_figure(image_path+'.pdf')
        self.canvas.print_figure(image_path+'.png')

        """
        except Exception as e:
            print "Exception in qkit/gui/plot/plot.py"
            print e
        """

    def plt_vector(self):
        """
        Plot one-dimensional dataset. Print data vs. x-coordinate.
        
        Args:
            self: Object of the h5plot class.
        Returns:
            No return variable. The function operates on the self- matplotlib 
            objects.
        """
        self.ax.set_title(self.hf._filename[:-3]+" "+str3(self.ds.attrs.get('name','_name_')))
        self.y_data = np.array(self.ds)
        self.y_exp = self._get_exp(self.y_data)
        self.y_label = concat(self.ds.attrs.get('name','_name_'),' (',self._unit_prefixes[self.y_exp],self.ds.attrs.get('unit','_unit_'),')')
        try:
            self.x_data = self.hf[self.x_ds_url]
            self.x_exp = self._get_exp(np.array(self.x_data))
            self.x_label = concat(self.x_data.attrs.get('name','_xname_'),' (', self._unit_prefixes[self.x_exp],self.x_data.attrs.get('unit','_xunit_'),')')
        except Exception:
            self.x_data = np.arange(len(self.y_data))
            self.x_label = '_none_ / _none_'
        plot_style = self.plot_styles[int(not (len(self.y_data) - 1)) * 2] # default is 'x' for one point and '-' for lines
        #if len(self.y_data) == 1: #only one entry, print as cross
        #    plot_style = 'x'
        #else:
        #    plot_style = '-'
        try:
            self.ax.plot(np.array(self.x_data)*10**int(-self.x_exp), self.y_data[0:len(self.x_data)]*10**int(-self.y_exp), plot_style)   #JB: avoid crash after pressing the stop button when arrays are of different lengths
        except TypeError:
            self.ax.plot(0,self.y_data, plot_style)

    def plt_matrix(self):
        """
        Plot two-dimensional dataset. print data color-coded y-coordinate
        vs. x-coordinate.
        
        Args:
            self: Object of the h5plot class.
        Returns:
            No return variable. The function operates on the self- matplotlib 
            objects.
        """
        self.x_ds = self.hf[self.x_ds_url]
        self.x_exp = self._get_exp(np.array(self.x_ds))
        self.x_label = concat(self.x_ds.attrs.get('name', '_xname_'),' (',self._unit_prefixes[self.x_exp],self.x_ds.attrs.get('unit','_xunit_'),')')
        self.y_ds = self.hf[self.y_ds_url]
        self.y_exp = self._get_exp(np.array(self.y_ds))
        self.y_label = concat(self.y_ds.attrs.get('name', '_yname_'),' (', self._unit_prefixes[self.y_exp] ,self.y_ds.attrs.get('unit','_yunit_'),')')
        self.ds_data = np.array(self.ds).T #transpose matrix to get x/y axis correct
        self.ds_exp = self._get_exp(self.ds_data)
        self.ds_data *= 10.**-self.ds_exp
        self.ds_label = concat(self.ds.attrs.get('name', '_name_'),' (',self._unit_prefixes[self.ds_exp],self.ds.attrs.get('unit', '_unit_'),')')

        x_data = np.array(self.x_ds)*10.**-self.x_exp
        x_min, x_max = np.amin(x_data), np.amax(x_data)
        dx = self.x_ds.attrs.get('dx', (x_data[-1]-x_data[0])/(len(x_data)-1))
        y_data = np.array(self.y_ds)*10.**-self.y_exp
        y_min, y_max = np.amin(y_data), np.amax(y_data)
        dy = self.y_ds.attrs.get('dy', (y_data[-1]-y_data[0])/(len(y_data)-1))

        # downsweeps in any direction have to be corrected
        # this is triggered by dx/dy values < 0
        # data-matrix and min/max-values have to be swapped
        if dx < 0:
            self.ds_data = np.fliplr(self.ds_data)
        if dy < 0:
            self.ds_data = np.flipud(self.ds_data)

        # plot
        self.ax.set_title(self.hf._filename[:-3]+" "+str3(self.ds.attrs.get('name','_name_')))
        self.cax = self.ax.imshow(self.ds_data,
                                  extent=(x_min, x_max, y_min, y_max),
                                  aspect='auto',
                                  origin='lower',
                                  vmin=self._get_vrange(self.ds_data, 2)[0], vmax=self._get_vrange(self.ds_data, 2)[1],
                                  interpolation='none')
        #self.cax = self.ax.pcolormesh(x_data, y_data, self.ds_data)
        self.cax.set_rasterized(True)
        self.cbar = self.fig.colorbar(self.cax)
        self.cbar.ax.set_ylabel(self.ds_label)
        self.cbar.ax.yaxis.label.set_fontsize(20)
        for i in self.cbar.ax.get_yticklabels():
            i.set_fontsize(16)

    def plt_box(self):
        """
        Plot two-dimensional dataset. Print data color-coded y-coordinate
        vs. x-coordinate.

        Args:
            self: Object of the h5plot class.
        Returns:
            No return variable. The function operates on the self- matplotlib
            objects.
        """
        self.x_ds = self.hf[self.x_ds_url]
        self.x_exp = self._get_exp(np.array(self.x_ds))
        self.x_label = concat(self.x_ds.attrs.get('name', '_xname_'),' (',self._unit_prefixes[self.x_exp],self.x_ds.attrs.get('unit','_xunit_'),')')
        self.y_ds = self.hf[self.y_ds_url]
        self.y_exp = self._get_exp(np.array(self.y_ds))
        self.y_label = concat(self.y_ds.attrs.get('name', '_yname_'),' (',self._unit_prefixes[self.y_exp],self.y_ds.attrs.get('unit', '_yunit_'),')')
        self.z_ds = self.hf[self.z_ds_url]
        self.z_exp = self._get_exp(np.array(self.z_ds))
        self.z_label = concat(self.z_ds.attrs.get('name', '_zname_'),' (',self._unit_prefixes[self.z_exp],self.z_ds.attrs.get('unit','_zunit_'), ')')
        self.ds_data = np.array(self.ds)[:, :, self.ds.shape[2] // 2].T  # transpose matrix to get x/y axis correct
        self.ds_exp = self._get_exp(self.ds_data)
        self.ds_data *= 10.**-self.ds_exp
        self.ds_label = concat(self.ds.attrs.get('name', '_name_'),' (',self._unit_prefixes[self.ds_exp],self.ds.attrs.get('unit','_unit_'),')')

        x_data = np.array(self.x_ds)*10.**-self.x_exp
        x_min, x_max = np.amin(x_data), np.amax(x_data)
        dx = self.x_ds.attrs.get('dx', (x_data[-1]-x_data[0])/(len(x_data)-1))
        y_data = np.array(self.y_ds)*10.**-self.y_exp
        y_min, y_max = np.amin(y_data), np.amax(y_data)
        dy = self.y_ds.attrs.get('dy', (y_data[-1]-y_data[0])/(len(y_data)-1))

        # downsweeps in any direction have to be corrected
        # this is triggered by dx/dy values < 0
        # data-matrix and min/max-values have to be swapped
        if dx < 0:
            self.ds_data = np.fliplr(self.ds_data)
        if dy < 0:
            self.ds_data = np.flipud(self.ds_data)

        # plot
        self.ax.set_title(concat(self.hf._filename[:-3]," ",self.ds.attrs.get('name','_name_')))
        self.cax = self.ax.imshow(self.ds_data,
                                  extent=(x_min, x_max, y_min, y_max),
                                  aspect='auto',
                                  origin='lower',
                                  vmin=self._get_vrange(self.ds_data, 2)[0], vmax=self._get_vrange(self.ds_data, 2)[1],
                                  interpolation='none')
        #self.cax = self.ax.pcolormesh(x_data, y_data, self.ds_data)
        self.cax.set_rasterized(True)
        self.cbar = self.fig.colorbar(self.cax)
        self.cbar.ax.set_ylabel(self.ds_label)
        self.cbar.ax.yaxis.label.set_fontsize(20)
        for i in self.cbar.ax.get_yticklabels():
            i.set_fontsize(16)

    def plt_coord(self):
        # not (yet?) implemented. we'll see ...
        pass

    def plt_txt(self):
        # not (yet?) implemented. we'll see ...
        pass

    def plt_view(self):
        """
        Plot views with possible multi-level overlays.
        
        First shot at (automatically) plotting views.
        Since this structure is rather flexible, there is no universal approach
        for meaningful plots. First demand was IV curves from the transport 
        measurements.
        The code is a recycled version of the _display_1D_view() fct of 
        qkit.gui.qviewkit.PlotWindow_lib
        
        Args:
            self: Object of the h5plot class.
        Returns:
            No return variable. The function operates on the self- matplotlib 
            objects.
        """
        # views are organized in overlays, the number of x vs y plot in one figure (i.e. data and fit)
        overlay_num = self.ds.attrs.get("overlays",0)
        overlay_urls = []
        err_urls = []
        self.ax.set_title(self.hf._filename[:-3]+" "+str3(self.ds.attrs.get('name','_name_')))
        
        # the overlay_urls (urls of the x and y datasets that ar plotted) are extracted from the metadata
        for i in range(overlay_num+1):
            ov = str3(self.ds.attrs.get("xy_"+str(i),""))
            if ov:
                overlay_urls.append(ov.split(":"))
            err_urls.append(self.ds.attrs.get("xy_"+str(i)+"_error",""))
                
        self.ds_xs = []
        self.ds_ys = []
        self.ds_errs = []
        for xy in overlay_urls:
            self.ds_xs.append(self.hf[xy[0]])
            self.ds_ys.append(self.hf[xy[1]])
            
        for err_url in err_urls:
            try:
                self.ds_errs.append(self.hf[err_url])
            except:
                self.ds_errs.append(0)
        
        """
        the ds_type are detected. this determines which dataset is displayed as a 1D plot.
        since the initial demand for the plotting comes from IV measurements and for easy handling,
        the default display in qviewkit is saved.
        """
        self.ds_label = str3(self.ds.attrs.get('name','_name_'))
        view_params = json.loads(str3(self.ds.attrs.get("view_params", {})))
        #if 'aspect' in view_params:
        #    self.ax.set_aspect = view_params.get('aspect', 'auto')
        markersize = view_params.get('markersize', 5)
        # iteratring over all x- (and y-)datasets and checking for dimensions gives the data to be plotted
        for i, x_ds in enumerate(self.ds_xs):
            y_ds = self.ds_ys[i]
            err_ds = self.ds_errs[i]
            
            #1D data is easy, for matrix and box the very last recorded 1D data is plotted
            if x_ds.attrs.get('ds_type',0) == ds_types['coordinate'] or x_ds.attrs.get('ds_type',0) == ds_types['vector']:
                if y_ds.attrs.get('ds_type',0) == ds_types['vector'] or y_ds.attrs.get('ds_type',0) == ds_types['coordinate']:
                    x_data = np.array(x_ds)
                    y_data = np.array(y_ds)
                    if err_ds:
                        err_data = np.array(err_ds)
    
                elif y_ds.attrs.get('ds_type',0) == ds_types['matrix']:
                    x_data = np.array(x_ds)
                    if "default_trace" in  view_params:
                        y_data = np.array(y_ds[:,view_params.get("default_trace",-1)])
                    else:
                        y_data = np.array(y_ds[-1]) # The code was like this, but I believe the axis is wrong
                    if err_ds:
                        err_data = np.array(err_ds[-1])
    
                elif y_ds.attrs.get('ds_type',0) == ds_types['box']:
                    x_data = np.array(x_ds)
                    y_data = np.array(y_ds[-1,-1,:])
                    if err_ds:
                        err_data = np.array(err_ds[-1,-1,:])
    
            ## This is in our case used so far only for IQ plots. The functionality derives from this application.
            elif x_ds.attrs.get('ds_type',0) == ds_types['matrix']:
                x_data = np.array(x_ds[-1])
                y_data = np.array(y_ds[-1])
            
            elif x_ds.attrs.get('ds_type',0) == ds_types['box']:
                x_data = np.array(x_ds[-1,-1,:])
                y_data = np.array(y_ds[-1,-1,:])

            plot_style = self.plot_styles[view_params.get('plot_style', int(not (len(y_data) - 1)) * 2)] # default is 'x' for one point and '-' for lines
            if err_ds:
                self.ax.errorbar(x_data, y_data[0:len(x_data)], yerr=err_data[0:len(x_data)], label=y_ds.name.split('/')[-1])
            else:
                try:
                    x_exp = self._get_exp(x_data) # caculate order of magnitude for unit-prefix
                    y_exp = self._get_exp(y_data[0:len(x_data)])
                    self.ax.plot(x_data*10**-x_exp, y_data[0:len(x_data)]*10**-y_exp, plot_style, label=y_ds.name.split('/')[-1])
                except TypeError:
                    self.ax.plot(0,y_data, plot_style)

            # x- and y-labels come from view_params['labels'] or if not provided from the last added entry in the overlay
            if view_params.get("labels", False):
                self.x_label = view_params['labels'][0]
                self.y_label = view_params['labels'][1]
            else:
                self.x_label = str3(x_ds.attrs.get("name", "_none_"))
                self.y_label = str3(y_ds.attrs.get("name", "_none_"))
            self.x_unit = str3(x_ds.attrs.get("unit","_none_"))
            self.y_unit = str3(y_ds.attrs.get("unit","_none_"))
            self.x_label += ' (' + self._unit_prefixes[x_exp] + self.x_unit + ')'
            self.y_label += ' (' + self._unit_prefixes[y_exp] + self.y_unit + ')'
        self.ax.legend()

    def _get_exp(self, data):
        """
        This function calculates the order of magnitude (exponent in steps of 3) to use for unit-prefix.
        """
        try:
            return np.nanmax(np.log10(np.abs(data[data != 0]))) // 3 * 3
        except:
            return 0

    def _get_vrange(self, data, percent):
        '''
        This function calculates ranges for the colorbar to get rid of spikes in the data.
        If the data is evenly distributed, this should not change anything in your colorbar.
        '''
        _min = np.nanpercentile(data,percent)
        _max = np.nanpercentile(data,100-percent)
        _min -= (_max-_min)*percent/(100.-2*percent)
        _max += (_max-_min)*percent/(100.-2*percent)
        return [_min,_max]


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(
    description="plot.py hdf and matplotlib-based plotting of datasets / KIT 2015")

    parser.add_argument('-f','--file',required=True,type=str, help='hdf/h5 filename to open')
    parser.add_argument('-c', '--comment', type=str, help='(optional) comment to append at filenames')
    parser.add_argument('-pdf','--save-pdf', default=False,action='store_true', help='(optional) save default plots')

    args=parser.parse_args()

    # get the full path
    filepath= os.path.abspath(args.file)
    save_plots(filepath,comment=args.comment,save_pdf=args.save_pdf)




SIGNALS = {
'info':         5,
'update':       10,
'reload':       11,
'stop-request': 12,
'close-gui':    13,
'changed':          14, # self._changed

'measurement-start':20,
'measurement-end':  21,
'measurement-idle': 22,
'new-data-point':   23,
'new-data-block':   24,

'removed':          30, 		# self.get_name()
'parameter-added':  32,	# name
'parameter-removed':33,	# name
'parameter-changed':34,	# name


'tags-added':       40, #newtags
'instrument-added': 41,	# name
'instrument-removed':42,# name
'instrument-changed':43,# sender.get_name(), changes
'item-added':       44, 	# name
}



"""
wafeform plotter JB@KIT 02/2016 jochen.braumueller@kit.edu

The waveform plotter wf_plot.py visualizes the waveform generated and employed for a measurement.
Prospectively, the plotter may be called by default to save a picture of the pulse sequence.

input:		analog_wfm: single (homodyne) or tuple of matrices carrying analog entries for traces and points in AWG window
            seq: sequence to be plotted, default 1
            complete_marker: single marker matrice or set of up to for marker lists (possibly arranged as [[...],[...]])
            xrange: xrange to be plotted
            sample
output: -
"""

import numpy as np
import matplotlib.pyplot as plt
# import os, glob
import time
import logging

no_qkit = False
try:
    import qkit
    data_dir_config = qkit.cfg.get('datadir')
except ImportError:
    logging.warning('no qkit environment')
    no_qkit = True


def wfplot(analog_wfm, seq=1, complete_marker=None, x_range=None, sample=None):
    font = {'weight': 'normal', 'size': 16}
    plt.rc('font', **font)
    labelsize = 16
    fig, axis = plt.subplots(figsize=(15, 5))

    # extract x axis information
    if sample != None:
        xval = np.linspace(0, sample.exc_T * 1e6, int(np.round(sample.exc_T * sample.clock) / 4) * 4)  # in us
        axis.set_xlabel(r'$time\,(\mathrm{\mu s})$', fontsize=labelsize)

        try:
            t = np.arange(0, sample.readout_tone_length, 1 / sample.readout_clock)
            r1 = np.sin(2 * np.pi * t * sample.readout_iq_frequency)
            r2 = np.cos(2 * np.pi * t * sample.readout_iq_frequency)
            axis.plot((sample.exc_T - sample.overlap * 1e-9) * 1e6 + range(len(r1)), r1, 'black')
            axis.plot((sample.exc_T - sample.overlap * 1e-9) * 1e6 + range(len(r1)), r2, 'black')
        except:
            logging.warning('readout pulse not plotted: sample parameters not found')
    else:
        axis.set_xlabel('# samples', fontsize=labelsize)
        if isinstance(analog_wfm[seq][0], (list, tuple, np.ndarray)):  # heterodyne mode
            xval = np.arange(0, len(analog_wfm[seq][0]))
        else:
            xval = np.arange(0, len(analog_wfm[seq]))

    if isinstance(analog_wfm[seq][0], (list, tuple, np.ndarray)):  # heterodyne mode
        axis.plot(xval, analog_wfm[seq][0], 'red', alpha=0.7, label='I')
        axis.plot(xval, analog_wfm[seq][1], 'blue', alpha=0.7, label='Q')
    else:
        axis.fill_between(xval, 0, analog_wfm[seq], color='red', alpha=0.7)
        axis.plot(xval, analog_wfm[seq], 'r-', alpha=0.7, label='homodyne')

    if complete_marker != None:
        clr_dict = {0: 'grey', 1: 'magneta', 2: 'green', 3: 'tan'}
        if len(np.array(complete_marker).shape) > 2:  # more than one single marker
            markers = np.array(complete_marker).reshape(4, np.array(complete_marker).shape[2],
                                                        np.array(complete_marker).shape[3])
        else:
            markers = [np.array(complete_marker)]

        maxis = axis.twinx()
        xlm = len(xval)
        xrm = 0
        for mi, m in enumerate(markers):
            if (m != np.zeros_like(m)).any() and m != None:
                maxis.fill_between(xval, 0, float(4 - mi) / 2 * m[seq], color=clr_dict[mi], alpha=0.7)
                maxis.plot(xval, float(4 - mi) / 2 * m[seq], color=clr_dict[mi], alpha=0.7, label='m%d' % mi)
                ''' find appropriate boundaries for the plot window '''
                xl = np.min(xlm, np.where(np.array(m[seq]) > 0)[0][0])
                xr = np.max(xrm, np.where(np.array(m[seq]) > 0)[0][-1])
        maxis.legend(loc=1)

    axis.legend(loc=2)
    if xrange != None:
        plt.xlim(x_range)
    else:
        ''' find appropriate window by cutting off edge regions filled with zeros '''
        if isinstance(analog_wfm[seq][0], (list, tuple, np.ndarray)):  # heterodyne mode
            xl = np.min(xl, np.where(np.gradient(analog_wfm[seq][0]) > 0)[0][0])
            xr = np.max(xr, np.where(np.gradient(analog_wfm[seq][0]) > 0)[0][-1])
        else:
            xl = np.min(xl, np.where(np.gradient(analog_wfm[seq]) > 0)[0][0])
            xr = np.max(xr, np.where(np.gradient(analog_wfm[seq]) > 0)[0][-1])

        # put some space left and right and check whether this is still a valid window
        xl -= 100
        if xl < 0:
            xl = 0
        xr += 100
        if xr > len(xval):
            xr = len(xval)

        plt.xlim([xl, xr])

    fig.tight_layout()



#
# This file handles external requests to the radiall switch pi
# S1@KIT2020
# Based on TIP 2.0 by HR@KIT 2019
#
import logging
import zmq
import zmq.auth
from switch import Switch


def _isfloat(argument):
    try:
        float(argument)
        return True
    except ValueError:
        return False


def parse_request(request, switch):
    """
    The following commands are available. They are not case sensitive and can be abbreviated up to the first letter. Sub-commands and parameters are
    separated by a slash.
    All switch and port numbering starts with 1, i.e. switches 1,2 ports 1-6
    
    SET/switch/port: e.g. SET/1/3 switches switch No 1 to Port 3 (and resets the other ports, if needed)
    
    ENABLE/switch/port: activate the specific port. This does not check wether you have multiple ports enabled simultaneously.
    DISABLE/switch/port: disables (resets) the specific port.
    
    GET/switch: returns a list of currently activated ports.
    GET/switch/port: returns True if the specific port is activated.
    
    RESET/switch: Perform a full reset of all ports (sequentially) for the given switch
    
    SET/LENGTH/length: e.g. SET/LENGTH/0.5 sets the current pulse time to 0.5s
    SET/LENGTH/length/switch: SET/LENGTH/0.2/1 sets pulse time to 0.2s for switch 1
    GET/LENGTH: returns the current pulse times
    
    """
    logging.info(request)
    
    cmds = request.strip('/').split("/")
    cmds = [cmd.strip().lower() for cmd in cmds]
    try:
        if "set".find(cmds[0]) == 0:
            if _isfloat(cmds[1]) and _isfloat(cmds[2]):
                if int(cmds[1]) in [1, 2] and int(cmds[2]) in [1, 2, 3, 4, 5, 6]:
                    switch.switch_to(int(cmds[1])-1, int(cmds[2])-1)
                    return True
                else:
                    return "Error: Argument 1 needs to be [1|2], argument 2 in [1,6]"
            elif "length".find(cmds[1]) == 0 and _isfloat(cmds[2]):
                if len(cmds) == 4 and _isfloat(cmds[3]) and int(cmds[3]) in [1,2]:
                    switch.set_pulse_length(length=float(cmds[2]),switch=int(cmds[3]))
                else:
                    switch.set_pulse_length(float(cmds[2]))
                return True
        elif "enable".find(cmds[0]) == 0 and _isfloat(cmds[1]) and _isfloat(cmds[2]):
            if int(cmds[1]) in [1, 2] and int(cmds[2]) in [1, 2, 3, 4, 5, 6]:
                switch.enable(int(cmds[1])-1, int(cmds[2])-1)
                return True
            else:
                return "Error: Argument 1 needs to be [1|2], argument 2 in [1,6]"
        elif "disable".find(cmds[0]) == 0 and _isfloat(cmds[1]) and _isfloat(cmds[2]):
            if int(cmds[1]) in [1, 2] and int(cmds[2]) in [1, 2, 3, 4, 5, 6]:
                switch.disable(int(cmds[1])-1, int(cmds[2])-1)
                return True
            else:
                return "Error: Argument 1 needs to be [1|2], argument 2 in [1,6]"
        elif "reset".find(cmds[0]) == 0:
            if len(cmds) == 1:
                switch.reset_all(None)
                return True
            elif int(cmds[1]) in [1, 2]:
                switch.reset_all(int(cmds[1])-1)
                return True
            else:
                return "Error: Argument 1 needs to be [1|2]"
        elif "get".find(cmds[0]) == 0:
            if _isfloat(cmds[1]):
                if len(cmds) == 2 and int(cmds[1]) in [1, 2]:
                    return [s+1 for s in switch.get_switch_position(int(cmds[1])-1)]
                elif int(cmds[1]) in [1, 2] and _isfloat(cmds[2]) and int(cmds[2]) in [1, 2, 3, 4, 5, 6]:
                    return int(cmds[2])-1 in switch.get_switch_position(int(cmds[1])-1)
                else:
                    return "Error: Argument 1 needs to be [1|2], argument 2 in [1,6] or not present"
            elif "length".find(cmds[1]) == 0:
                return switch.get_pulse_length()
        else:
            return "invalid syntax : " + request
        return "Request not processed: " + request
    except Exception as e:
        return "Command error. Your command '" + request + "'. Error: '" + str(e) + "'."


def serve_requests():
    context = zmq.Context()
    
    # # FIXME: authentication is not working in the moment
    # auth = ThreadAuthenticator(context, log=logging.getLogger())
    # auth.start()
    # auth.allow('127.0.0.1')
    # auth.allow('localhost')
    # set_allowed_IPs(auth)
    
    socket = context.socket(zmq.REP)
    socket.zap_domain = b'global'
    socket.bind("tcp://*:5000")
    
    switch = Switch()
    
    while True:
        #  Wait for next request from client
        message = socket.recv_string()
        logging.debug("Received request: %s" % message)
        socket.send_string(str(parse_request(message, switch)))

# def set_allowed_IPs(auth):
#     for IP in str(config['system'].get('allowed_ips', "")).split(" "):
#         logging.info("Add " + IP + " to allowed IP list.")
#         auth.allow(IP)

if __name__ == "__main__":
    print("Switch server starting")
    serve_requests()



from qkit.core.instrument_base import Instrument

class virtual_step_attenuator(Instrument):
    '''
        this is real hardware connected to the experiment
        add one of those to automatically note its value in the measurement files
    '''

    def __init__(self, name):
        Instrument.__init__(self, name, tags=['virtual'])
        
        self.add_parameter('attenuation', type=float,
            flags=Instrument.FLAG_SET, units='dB')

    def do_set_attenuation(self, attn):
        self._attn = attn

    def do_get_attenuation(self):
        return self._attn



# -*- coding: utf-8 -*-
"""
Created 2015

@author: hrotzing

Modified 2017 by S1
"""
import logging
import os
import time
import qkit

alphabet = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"


class DateTimeGenerator(object):
    """DateTimeGenerator class to provide a timestamp for each measurement file
    upon creation.
    
    For sorting and unique identification across multiple measurement setups
    we provide a timestamp for each h5 file created with qkit. We use the unix
    timestamp to get time resolution of one second which should be enough for
    our needs. The integer timestamp is then converted either into a HHMMSS
    representation using day and month information to create a folder, or the
    timestamp gets converted using the alphabet to create a 6 digit UUID.
    """
    
    def __init__(self):
        self.returndict = {}
        self.returndict['_unix_timestamp'] = int(time.time())
        self.returndict['_localtime'] = time.localtime(self.returndict['_unix_timestamp'])
        self.returndict['_timestamp'] = time.asctime(self.returndict['_localtime'])
        self.returndict['_timemark'] = time.strftime('%H%M%S', self.returndict['_localtime'])
        self.returndict['_datemark'] = time.strftime('%Y%m%d', self.returndict['_localtime'])
        self.returndict['_uuid'] = encode_uuid(self.returndict['_unix_timestamp'])
    
    # call for h5 filename, qkit config gets checked for encoding procedure.
    def new_filename(self, name=None):
        if qkit.cfg.get('datafolder_structure',1) == 2:
            self.new_filename_v2(name) # 6 digit UUID
        else:
            self.new_filename_v1(name) # HHMMSS representation
        self.returndict['_folder'] = os.path.join(qkit.cfg['datadir'], self.returndict['_relfolder'])
        self.returndict['_relpath'] = os.path.join(self.returndict['_relfolder'],self.returndict['_filename'])
        self.returndict['_filepath'] = os.path.join(self.returndict['_folder'], self.returndict['_filename'])
        
        return self.returndict
    
    def new_filename_v1(self, name):
        filename = str(self.returndict['_timemark'])
        if name != '' and name is not None:
            filename += '_' + str(name)
        self.returndict['_filename'] = filename + '.h5'
        '''Old filename with datadir/YYMMDD/HHMMSS_name/HHMMSS_name.h5'''
        self.returndict['_relfolder'] = os.path.join(
                self.returndict['_datemark'],
                filename
        )
    
    def new_filename_v2(self, name):
        filename = str(self.returndict['_uuid'])
        if name != '' and name is not None:
            filename += '_' + str(name)
        self.returndict['_filename'] = filename + '.h5'
        '''New filename with datadir/run_id/user/uuid_name/uuid_name.h5'''
        
        self.returndict['_relfolder'] = os.path.join(
                qkit.cfg.get('run_id', 'NO_RUN').strip().replace(" ", "_").upper(),
                qkit.cfg.get('user', 'John_Doe').strip().replace(" ", "_"),
                filename
        )


def encode_uuid(value):
    """Encodes the integer unix timestamp into a 6 digit UUID using the alphabet.
    
    Args:
        Integer-cast unix timestamp.
    Return:
        6 digit UUID string.
    """
    # if not value: value = self._unix_timestamp
    output = ''
    la = len(alphabet)
    while value:
        output += alphabet[value % la]
        value = int(value / la)
    return output[::-1]


def decode_uuid(string):
    """Decodes the 6 digit UUID back into integer unix timestamp.
    
    Args:
        6 digit UUID string.        
    Return:
        Integer-cast unix timestamp.
    """
    # if not string: string = self._uuid
    output = 0
    multiplier = 1
    string = string[::-1].upper()
    la = len(alphabet)
    while string != '':
        f = alphabet.find(string[0])
        if f == -1:
            raise ValueError("Can not decode this: {}<--".format(string[::-1]))
        output += f * multiplier
        multiplier *= la
        string = string[1:]
    return output



# -*- coding: utf-8 -*-
"""
@author: S1@KIT/2018
This file checks if we are in a git environment and provides information about the last commit
"""
import logging
from os.path import join
from time import time

import qkit
if qkit.cfg.get('check_for_updates',False):
    try:
        with open(join(qkit.cfg['qkitdir'], '../.git/logs/HEAD'), 'rb') as f:
            f.seek(-1024, 2)
            last_commit = f.readlines()[-1].decode().split("\t")[0]  # The line looks like this: 0old_commit 1new_commit 2username 3mail 4timestamp 5timezone
            qkit.git = {
                'timestamp': float(last_commit.split(" ")[-2]),
                'commit_id': last_commit.split(" ")[1]
            }
            if (time() - qkit.git['timestamp']) / 3600 / 24 > 21:  # If the last commit is older than 3 weeks
                logging.warning("Your qkit version is older than 3 weeks. Please consider the 'git pull' command. We are usually trying to get better.")
    except Exception:
        qkit.git = {
            'timestamp': None,
            'commit_id': "UNTRACKED-NO_GIT_FOUND"
        }
        logging.info("You are not operating qkit from a git repository. You do not need to do so, but this is the easiest way to get always the latest version.")
else:
    qkit.cfg['check_for_updates'] = False
    logging.info("Not checking for updates.")


# -*- coding: utf-8 -*-

# Keithley.py driver for Keithley 2636A multi channel source measure unit
# Micha Wildermuth, micha.wildermuth@kit.edu 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import qkit
from qkit.core.instrument_base import Instrument
import logging
import numpy as np


class virtual_tunnel_electronic(Instrument):
    """
    This is the virtual driver for the home made tunnel electronic combined with a Multi Channel Source Measure Unit <SMU>. I.a. it allows the external usage of voltage controlled current sources or transimpedance amplifiers

    Usage:
    Initialize with
    <name> = qkit.instruments.create('<name>', 'virtual_tunnel_electronic', SMU=<SMU>)
    """

    def __init__(self, name, SMU):
        """
        Initializes VISA communication with the instrument Yokogawa GS820.

        Parameters
        ----------
        name: string
            Name of the instrument (driver).
        SMU: qkit.instrument
            Multi channel source measure unit.
        reset: bool, optional
            Resets the instrument to default conditions.. Default is False.

        Returns
        -------
        None

        Examples
        --------
        >>> import qkit
        QKIT configuration initialized -> available as qkit.cfg[...]

        >>> qkit.start()
        Starting QKIT framework ... -> qkit.core.startup
        Loading module ... S10_logging.py
        Loading module ... S12_lockfile.py
        Loading module ... S14_setup_directories.py
        Loading module ... S20_check_for_updates.py
        Loading module ... S25_info_service.py
        Loading module ... S30_qkit_start.py
        Loading module ... S65_load_RI_service.py
        Loading module ... S70_load_visa.py
        Loading module ... S80_load_file_service.py
        Loading module ... S85_init_measurement.py
        Loading module ... S98_started.py
        Loading module ... S99_init_user.py

        >>> SMU = qkit.instruments.create('SMU', 'Keithley', address='TCPIP0::00.00.000.00::INSTR', reset=True)
        >>> IVD = qkit.instruments.create('IVD', 'virtual_tunnel_electronic', SMU=SMU)
        Initialized the file info database (qkit.fid) in 0.000 seconds.
        """
        self.__name__ = __name__
        # create instrument
        logging.info(__name__ + ': Initializing instrument virtual tunnel electronic')
        Instrument.__init__(self, name, tags=['virtual'])
        self._instruments = qkit.instruments.get_instruments()
        # Source Measure Unit (SMU)
        self._SMU = SMU
        # external measurement setup
        self._BW = None  # bandwidth of prefilter (only for get_all and setting file)
        self._dAdV = 2e-6  # for external current bias
        self._amp = 1e3  # for external voltage amplifier
        self._tau_amp = 100e-3  # time constant of amplifier (only for get_all and setting file)
        self._dVdA = 1e8  # for external voltage bias
        self._Vdiv = 1e3  # for external voltage divider
        # internal variables
        self._sweep_mode = 0  # VV-mode
        self._sweep_channels = (1, 2)
        self._sweep_modes = {0: 'VV-mode', 1: 'IV-mode', 2: 'VI-mode'}
        self._pseudo_bias_mode = 0  # current bias
        self._pseudo_bias_modes = {0: 'current bias', 1: 'voltage bias'}
        # dict of defaults values: defaults[<pseudo_bias_mode>][<parameter>][<value>]
        self._defaults = {0: {'dAdV': 2e-6,
                              'amp': 1e3},
                          1: {'dVdA': 1e8,
                              'Vdiv': 1e3}}
        self._default_str = 'default parameters'
        self.__set_defaults_docstring()

    def set_BW(self, val=None):
        """
        Sets the internal variable for the bandwidth of an external filter to <val>.

        Parameters
        ----------
        val: float
            Bandwidth of external filter (in Hz). Default is None.

        Returns
        -------
        None
        """
        self._BW = val
        return

    def get_BW(self):
        """
        Gets the internal variable for the bandwidth of an external filter.

        Parameters
        ----------
        None

        Returns
        -------
        val: float
            Bandwidth of external filter (in Hz). Default is None.
        """
        return self._BW

    def set_dAdV(self, val=1):
        """
        Sets the internal variable for the voltage-current conversion factor to <val>.
        This is used for the current bias if the currents to apply are generated by means of an external voltage controlled current source.

        Parameters
        ----------
        val: float
            Voltage-current conversion factor (in A/V). Default is 1.

        Returns
        -------
        None
        """
        self._dAdV = val
        return

    def get_dAdV(self):
        """
        Gets the internal variable for the voltage-current conversion factor.
        This is used for the current bias if the currents to apply are generated by means of an external voltage controlled current source.

        Parameters
        ----------
        None

        Returns
        -------
        val: float
            Voltage-current conversion factor (in A/V).
        """
        return self._dAdV

    def set_amp(self, val=1):
        """
        Sets the internal variable for the pre-amplification factor to <val>.
        This is used in the current bias if the measured voltages are amplified by means of an external pre-amplifier.

        Parameters
        ----------
        val: float
            Pre-amplification factor. Default is 1.

        Returns
        -------
        None
        """
        self._amp = val
        return

    def get_amp(self):
        """
        Gets the internal variable for the pre-amplification factor.
        This is used in the current bias if the measured voltages are amplified by means of an external pre-amplifier.

        Parameters
        ----------
        None

        Returns
        -------
        val: float
            Pre-amplification factor.
        """
        return self._amp

    def set_tau_amp(self, val=100e-3):
        """
        Sets the internal variable for the pre-amplifiers lowpass filters time constant to <val>.
        This is used in the current bias if the measured voltages are amplified by means of an external pre-amplifier whose feedback includes a RC element to lowpass filter the signal.

        Parameters
        ----------
        val: float
            Pre-amplifiers time constant (τ=2πRC in s). Default is 100e-3.

        Returns
        -------
        None
        """
        self._tau_amp = val
        return

    def get_tau_amp(self):
        """
        Gets the internal variable for the pre-amplifiers lowpass filters time constant.
        This is used in the current bias if the measured voltages are amplified by means of an external pre-amplifier whose feedback includes a RC element to lowpass filter the signal.

        Parameters
        ----------
        None

        Returns
        -------
        val: float
            Pre-amplifiers time constant (τ=2πRC in s).
        """
        return self._tau_amp

    def set_dVdA(self, val=1):
        """
        Sets the internal variable for the current-voltage conversion to <val>.
        This is used for the voltage bias if the measured currents are converted to voltages by means of an external transimpedance amplifier.

        Parameters
        ----------
        val: float
            Current-voltage conversion factor (in V/A). Default is 1.

        Returns
        -------
        None
        """
        self._dVdA = val

    def get_dVdA(self):
        """
        Gets the internal variable for the current-voltage conversion.
        This is used for the voltage bias if the measured currents are converted to voltages by means of an external transimpedance amplifier.

        Parameters
        ----------
        None

        Returns
        -------
        val: float
            Current-voltage conversion factor (in V/A).
        """
        return self._dVdA

    def set_Vdiv(self, val=1):
        """
        Sets the internal variable for the voltage division factor to <val>.
        This is used for the voltage bias if the input voltage is divided by means of an external voltage divider.

        Parameters
        ----------
        val: float
            Voltage division factor. Default is 1.

        Returns
        -------
        None
        """
        self._Vdiv = val

    def get_Vdiv(self):
        """
        Gets the internal variable for the voltage division factor.
        This is used for the voltage bias if the input voltage is divided by means of an external voltage divider.

        Parameters
        ----------
        None

        Returns
        -------
        val: float
            Voltage division factor.
        """
        return self._Vdiv

    def set_measurement_mode(self, mode, channel=1):
        """
        Sets measurement mode (wiring system) of channel <channel> to <mode>.

        Parameters
        ----------
        mode: int
            State of the measurement sense mode. Must be 0 (2-wire), 1 (4-wire) or 3 (calibration).
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        None
        """
        return self._SMU.set_measurement_mode(mode=mode, channel=channel)

    def get_measurement_mode(self, channel=1):
        """
        Gets measurement mode (wiring system) <mode> of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        mode: int
            State of the measurement sense mode. Meanings are 0 (2-wire), 1 (4-wire) and 3 (calibration).
        """
        return self._SMU.get_measurement_mode(channel=channel)

    def set_bias_mode(self, mode, channel=1):
        """
        Gets bias mode <mode> of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        mode: int
            Bias mode. Meanings are 0 (current) and 1 (voltage).
        """
        return self._SMU.set_bias_mode(mode=mode, channel=channel)

    def get_bias_mode(self, channel=1):
        """
        Sets sense mode of channel <channel> to mode <mode>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        mode: int
            Sense mode. Must be 0 (current), 1 (voltage), 2 (resistance) or 3 (power).

        Returns
        -------
            None
        """
        return self._SMU.get_bias_mode(channel=channel)

    def set_sense_mode(self, mode, channel=1):
        """
        Sets sense mode of channel <channel> to mode <mode>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        mode: int
            Sense mode. Must be 0 (current), 1 (voltage), 2 (resistance) or 3 (power).

        Returns
        -------
            None
        """
        return self._SMU.set_sense_mode(mode=mode, channel=channel)

    def get_sense_mode(self, channel=1):
        """
        Gets sense mode <mode> of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        mode: int
            Sense mode. Meanings are 0 (current), 1 (voltage), 2 (resistance) and 3 (power).
        """
        return self._SMU.get_sense_mode(channel=channel)

    def set_bias_range(self, val, channel=1):
        """
        Sets bias range of channel <channel> to <val>.

        Parameters
        ----------
        val: float
            Bias range. Possible values are -1 (auto), for currents 100pA, 1nA, 10nA, 100nA, 1uA, 10uA, 100uA, 1mA, 10mA, 100mA, 1 A, 1.5A and for voltages 200mV, 2V, 20V, 200V.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
            None
        """
        return self._SMU.set_bias_range(val=val, channel=channel)

    def get_bias_range(self, channel=1):
        """
        Gets bias mode <val> of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        val: float
            Bias range. Possible values for currents are 100pA, 1nA, 10nA, 100nA, 1uA, 10uA, 100uA, 1mA, 10mA, 100mA, 1 A, 1.5A and for voltages 200mV, 2V, 20V, 200V.
        """
        return self._SMU.get_bias_range(channel=channel)

    def set_sense_range(self, val, channel=1):
        """
        Sets sense range of channel <channel> to <val>.

        Parameters
        ----------
        val: float
            Sense range. Possible values are -1 (auto), for currents 100pA, 1nA, 10nA |100nA, 1uA, 10uA, 100uA, 1mA, 10mA, 100mA, 1 A, 1.5A and for voltages 200mV, 2V, 20V, 200V.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
            None
        """
        return self._SMU.set_sense_range(val=val, channel=channel)

    def get_sense_range(self, channel=1):
        """
        Gets sense mode <val> of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        val: float
            Sense range. Possible values are -1 (auto), for currents 100pA, 1nA, 10nA |100nA, 1uA, 10uA, 100uA, 1mA, 10mA, 100mA, 1 A, 1.5A and for voltages 200mV, 2V, 20V, 200V.
        """
        return self._SMU.get_sense_range(channel=channel)

    def set_bias_delay(self, val, channel=1):
        """
        Sets bias delay of channel <channel> to <val>.

        Parameters
        ----------
        val: float
            Bias delay with respect to the bias trigger. Possible values are -1 (auto), 0 (off) or positive numbers.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
            None
        """
        return self._SMU.set_bias_delay(val=val, channel=channel)

    def get_bias_delay(self, channel=1):
        """
        Gets bias delay <val> of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        val: float
            Bias delay with respect to the bias trigger.
        """
        return self._SMU.get_bias_delay(channel=channel)

    def set_sense_delay(self, val, channel=1, **kwargs):
        """
        Sets sense delay of channel <channel> to <val>.

        Parameters
        ----------
        val: float
            Sense delay with respect to the sense trigger. Possible values are -1 (auto), 0 (off) or positive numbers.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        factor: float, optional
            Multiplier to the delays if auto delay is used (for Keithley 2636A only). Default is 1.

        Returns
        -------
        None
        """
        return self._SMU.set_sense_delay(val=val, channel=channel, **kwargs)

    def get_sense_delay(self, channel=1):
        """
        Gets sense delay <val> of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        val: float
            Sense delay with respect to the sense trigger.
        """
        return self._SMU.get_sense_delay(channel=channel)

    def set_sense_average(self, val, channel=1, **kwargs):
        """
        Sets sense average of channel <channel> to <val>.

        Parameters
        ----------
        val: int
            Number of measured readings that are required to yield one filtered measurement. Must be in [1, 100].
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        mode: str
            Type of filter used for measurements when the measurement filter is enabled (for Keithley 2636A only). Possible values are 0 (moving average) | 1 (repeat average) | 2 (median). Default is 1.

        Returns
        -------
        None
        """
        return self._SMU.set_sense_average(val=val, channel=channel, **kwargs)

    def get_sense_average(self, channel=1):
        """
        Gets sense average status <status>, value <val> and mode <mode> of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        status: bool
            Status if the measurement filter is enabled.
        val: int
            Number of measured readings that are required to yield one filtered measurement.
        mode: str, optional
            Type of filter used for measurements when the measurement filter is enabled. Measnings are 0 (moving average) | 1 (repeat average) | 2 (median).
        """
        return self._SMU.get_sense_average(channel=channel)

    def set_plc(self, val):
        """
        Sets power line cycle (PLC) to <val>.

        Parameters
        ----------
        plc: int
            Power line frequency setting used for NPLC calculations. Possible values are -1 (auto) | 50 | 60.

        Returns
        -------
        None
        """
        return self._SMU.set_plc(val=val)

    def get_plc(self):
        """
        Gets power line cycle (PLC) <val>.

        Parameters
        ----------
        None

        Returns
        -------
        plc: int
            Power line frequency setting used for NPLC calculations.
        """
        return self._SMU.get_plc()

    def set_sense_nplc(self, val, channel=1):
        """
        Sets sense nplc (number of power line cycle) of channel <channel> with the <val>-fold of one power line cycle.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        val: float
            Integration aperture for measurements. Possible values depend on the used SMU.

        Returns
        -------
        None
        """
        return self._SMU.set_sense_nplc(val=val, channel=channel)

    def get_sense_nplc(self, channel=1):
        """
        Gets sense nplc (number of power line cycle) <val> of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        val: float
            Integration aperture for measurements.
        """
        return self._SMU.get_sense_nplc(channel=channel)

    def set_sense_autozero(self, val, channel=1):
        """
        Sets autozero of channel <channel> to <val>. Note that "on" means that the internal zero point is measured for each measurement, wherefore the measurement takes approximately twice as long as when the auto zero function is "off"

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        val: int
            Status of the internal reference measurements (autozero) of the source-measure unit. Possible values are 0 (off), 1 (on) or 2 (once).

        Returns
        -------
        None
        """
        return self._SMU.set_sense_autozero(val=val, channel=channel)

    def get_sense_autozero(self, channel=1):
        """
        Gets sense autozero <val> of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        val: int
            Status of the internal reference measurements (autozero) of the source-measure unit. Meanings are 0 (off), 1 (on) and 2 (once).
        """
        return self._SMU.get_sense_autozero(channel=channel)

    def set_status(self, status, channel=1):
        """
        Sets output status of channel <channel> to <status>.

        Parameters
        ----------
        status: int
            Output status. Possible values are 0 (off), 1 (on) or 2 (high Z).
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        None
        """
        return self._SMU.set_status(status=status, channel=channel)

    def get_status(self, channel=1):
        """
        Gets output status <status> of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        status: int
            Output status. Meanings are 0 (off), 1 (on) and 2 (high Z).
        """
        return self._SMU.get_status(channel=channel)

    def set_sweep_mode(self, mode=1, *channels):
        """
        Sets an variable to decide weather
         * voltage is both applied and measured (VV-mode),
         * current is applied and voltage is measured (IV-mode) or
         * voltage is applied and current is measured (VI-mode)
        and channels <channels> that are used during sweeps.

        Parameters
        ----------
        mode: int
            Sweep mode denoting bias and sense modes. Possible values are 0 (VV-mode), 1 (IV-mode) or 2 (VI-mode). Default is 1
        channels: int or tuple(int)
            Number of channel of usage. Must be 1 or 2 for IV-mode and VI-mode or a tuple containing both channels for VV-mode (1st argument as bias channel and 2nd argument as sense channel).

        Returns
        -------
        None
        """
        default_channels = {0: (1, 2), 1: (1,), 2: (1,)}
        if mode in [0, 1, 2]:
            if len(channels) == 0:
                logging.debug('{!s}: Set sweep channels to {!s}'.format(__name__, default_channels[mode]))
                self._sweep_channels = default_channels[mode]
                self._SMU._sweep_channels = default_channels[mode]
            else:
                logging.debug('{!s}: Set sweep channels to {!s}'.format(__name__, channels))
                self._sweep_channels = channels
                self._SMU._sweep_channels = channels
            logging.debug('{!s}: Set sweep mode to {:d}'.format(__name__, mode))
            self._sweep_mode = mode
            self._SMU.set_sweep_mode(mode=mode)
        else:
            logging.error('{!s}: Cannot set sweep mode to {!s}'.format(__name__, mode))
            raise ValueError('{!s}: Cannot set sweep mode to {!s}'.format(__name__, mode))

    def get_sweep_mode(self):
        """
        Gets an internal variable <mode> to decide weather
         * voltage is both applied and measured (VV-mode),
         * current is applied and voltage is measured (IV-mode) or
         * voltage is applied and current is measured (VI-mode).

        Parameters
        ----------
        None

        Returns
        -------
        mode: int
            Sweep mode denoting bias and sense modes. Meanings are 0 (VV-mode), 1 (IV-mode) or 2 (VI-mode).
        """
        return self._sweep_mode

    def get_sweep_channels(self):
        """
        Gets channels <channels> that are used during sweeps. In the IV and VI sweep mode one channel requires for biasing and sensing, whereas two channels are returned in the VV sweep mode (1st as bias and 2nd as sense channel).

        Parameters
        ----------
        None

        Returns
        -------
        channels: int or tuple(int)
            Number of channel of usage. Meanings are 1 or 2 for IV-mode and VI-mode or a tuple containing both channels for VV-mode (1st argument as bias channel and 2nd argument as sense channel).
        """
        if self._sweep_channels != self._SMU.get_sweep_channels():
            logging.error(
                '{!s}: sweep channels of {:s} and {:s} do not coincide: {:s} and {:s}'.format(__name__, self.__name__,
                                                                                              self._SMU.__name__,
                                                                                              self._sweep_channels,
                                                                                              self._SMU.get_sweep_channels()))
            raise ValueError(
                '{!s}: sweep channels of {:s} and {:s} do not coincide: {:s} and {:s}'.format(__name__, self.__name__,
                                                                                              self._SMU.__name__,
                                                                                              self._sweep_channels,
                                                                                              self._SMU.get_sweep_channels()))
        else:
            return self._sweep_channels

    def set_pseudo_bias_mode(self, mode):
        """
        Sets an internal variable according to the external measurement setup that proviedes an effective bias to <mode>.
         * For the current bias the exernal setup includes
            * a voltage-controlled current source to generate the bias currents and
            * a voltage pre-amplifier to amplify the measured voltages.
         * For the voltage bias the exernal setup includes
            * a voltage divider to reduce the bias voltages and
            * a transimpedance amplifier that converts the measured currents to voltages.

        Parameters
        ----------
        mode: int
            Pseudo bias mode denoting current and voltage bias. Possible values are 0 (current bias) and 1 (voltage bias).

        Returns
        -------
        None
        """
        if mode in [0, 1, 2]:
            logging.debug('{!s}: Set pseudo bias mode to {:d}'.format(__name__, mode))
            self._pseudo_bias_mode = mode
            self.__set_defaults_docstring()
        else:
            logging.error('{!s}: Cannot set pseudo bias mode to {!s}'.format(__name__, mode))
            raise ValueError('{!s}: Cannot set pseudo bias mode to {!s}'.format(__name__, mode))

    def get_pseudo_bias_mode(self):
        """
        Gets an internal variable according to the external measurement setup that proviedes an effective bias.
         * For the current bias the exernal setup includes
            * a voltage-controlled current source to generate the bias currents and
            * a voltage pre-amplifier to amplify the measured voltages.
         * For the voltage bias the exernal setup includes
            * a voltage divider to reduce the bias voltages and
            * a transimpedance amplifier that converts the measured currents to voltages.

        Parameters
        ----------
        None

        Returns
        -------
        mode: int
            Pseudo bias mode denoting current and voltage bias. Possible values are 0 (current bias) and 1 (voltage bias).
        """
        return self._pseudo_bias_mode

    def get_sweep_bias(self):
        """
        Gets the real bias mode as combination of <self._sweep_mode> and <self._pseudo_bias_mode>.

        Parameters
        ----------
        None

        Returns
        -------
        mode: int
            Bias mode. Meanings are 0 (current) and 1 (voltage).
        """
        # attribute **kwargs needed for identical call as in SMUs (which need channels)
        if self._sweep_mode != self._SMU.get_sweep_mode():
            logging.error(
                '{!s}: sweep mode of {:s} and {:s} do not coincide: {:s} and {:s}'.format(__name__, self.__name__,
                                                                                          self._SMU.__name__,
                                                                                          self._sweep_modes[
                                                                                              self._sweep_mode],
                                                                                          self._sweep_modes[
                                                                                              self._SMU.get_sweep_mode()]))
            raise ValueError(
                '{!s}: sweep mode of {:s} and {:s} do not coincide: {:s} and {:s}'.format(__name__, self.__name__,
                                                                                          self._SMU.__name__,
                                                                                          self._sweep_modes[
                                                                                              self._sweep_mode],
                                                                                          self._sweep_modes[
                                                                                              self._SMU.get_sweep_mode()]))
        else:
            return int(not bool(self._sweep_mode)) * self._pseudo_bias_mode + int(bool(self._sweep_mode)) * (
                        self._sweep_mode - 1)  # 0 (current bias) | 1 (voltage bias)

    def set_voltage(self, val, channel=1):
        """
        Sets bias voltage value of channel <channel> to <val>.
        This can only be set in the effective voltage bias mode, defined by <self.sweep_mode> and <self.pseudo_bias_mode>, and considers an external voltage divider <self._Vdiv>.

        Parameters
        ----------
        val: float
            Bias voltage value.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        None
        """
        if self._sweep_mode == 0:  # 0 (VV mode)
            if not self._pseudo_bias_mode:  # 0 (current bias)
                logging.error(
                    '{!s}: Cannot set voltage value of channel {!s} in the current bias'.format(__name__, channel))
                raise SystemError(
                    '{!s}: Cannot set voltage value of channel {!s} in the current bias'.format(__name__, channel))
            elif self._pseudo_bias_mode:  # 1 (voltage bias)
                return self._SMU.set_bias_value(val=val * self._Vdiv, channel=channel)
        elif self._sweep_mode == 1:  # 1 (IV mode)
            logging.error(
                '{!s}: Cannot set voltage value of channel {!s} in the current bias'.format(__name__, channel))
            raise SystemError(
                '{!s}: Cannot set voltage value of channel {!s} in the current bias'.format(__name__, channel))
        elif self._sweep_mode == 2:  # 2 (VI-mode)
            return self._SMU.set_bias_value(val=val, channel=channel)

    def get_voltage(self, channel=1):
        """
        Gets voltage value of channel <channel> taking tunnel settings of the electronic into account (<sweep_mode>, <pseudo_bias_mode>)

        Gets bias or sense voltage value of channel <channel>.
        This depends on the effective bias mode, defined by <self.sweep_mode> and <self.pseudo_bias_mode>, and considers the external pre-amplifier <self._amp> (current bias) and the external voltage divider <self._Vdiv> (voltage bias).

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        val: float
            Bias or sense voltage value.
        """
        if self._sweep_mode == 0:  # 0 (VV mode)
            if not self._pseudo_bias_mode:  # 0 (current bias)
                return self._SMU.get_sense_value(channel=channel) / self._amp
            elif self._pseudo_bias_mode:  # 1 (voltage bias)
                return self._SMU.get_bias_value(channel=channel) / self._Vdiv
        elif self._sweep_mode == 1:  # 1 (IV mode)
            return self._SMU.get_sense_value(channel=channel)
        elif self._sweep_mode == 2:  # 2 (VI-mode)
            return self._SMU.get_bias_value(channel=channel)

    def set_current(self, val, channel=1):
        """
        Sets bias current value of channel <channel> to <val>.
        This can only be set in the effective current bias mode, defined by <self.sweep_mode> and <self.pseudo_bias_mode>, and considers an external voltage-controlled current source <self._dAdV>.

        Parameters
        ----------
        val: float
            Bias current value.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        None
        """
        if self._sweep_mode == 0:  # 0 (VV mode)
            if not self._pseudo_bias_mode:  # 0 (current bias)
                return self._SMU.set_bias_value(val=val / self._dAdV, channel=channel)
            elif self._pseudo_bias_mode:  # 1 (voltage bias)
                logging.error(
                    '{!s}: Cannot set current value of channel {!s} in the voltage bias'.format(__name__, channel))
                raise SystemError(
                    '{!s}: Cannot set current value of channel {!s} in the voltage bias'.format(__name__, channel))
        elif self._sweep_mode == 1:  # 1 (IV mode)
            return self._SMU.set_bias_value(val=val, channel=channel)
        elif self._sweep_mode == 2:  # 2 (VI-mode)
            logging.error(
                '{!s}: Cannot set current value of channel {!s} in the voltage bias'.format(__name__, channel))
            raise SystemError(
                '{!s}: Cannot set current value of channel {!s} in the voltage bias'.format(__name__, channel))

    def get_current(self, channel=1):
        """
        Gets current value of channel <channel> taking tunnel settings of the electronic into account (<sweep_mode>, <pseudo_bias_mode>)

        Gets bias or sense current value of channel <channel>.
        This depends on the effective bias mode, defined by <self.sweep_mode> and <self.pseudo_bias_mode>, and considers the external voltage-controlled current source <self._dAdV> (current bias) and the external transimpedance amplifier <self._dVdA> (voltage bias).

        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        val: float
            Bias or sense current value.
        """
        if self._sweep_mode == 0:  # 0 (VV mode)
            if not self._pseudo_bias_mode:  # 0 (current bias)
                return self._SMU.get_bias_value(channel=channel) * self._dAdV
            elif self._pseudo_bias_mode:  # 1 (voltage bias)
                return self._SMU.get_sense_value(channel=channel) / self._dVdA
        elif self._sweep_mode == 1:  # 1 (IV mode)
            return self._SMU.get_bias_value(channel=channel)
        elif self._sweep_mode == 2:  # 2 (VI-mode)
            return self._SMU.get_sense_value(channel=channel)

    def ramp_voltage(self, stop, step, step_time=0.1, channel=1):
        """
        Ramps voltage of channel <channel> from recent value to stop value <stop> with step size <step> and step time <step_time> according to bias_mode.
        This can only be set in the effective voltage bias mode, defined by <self.sweep_mode> and <self.pseudo_bias_mode>, and considers an external voltage divider <self._Vdiv>.

        Parameters
        ----------
        stop: float
            Stop voltage value
        step: float
            Step voltage size.
        step_time: float
            Sleep time between voltage staircase ramp.
        channel: int
            Number of channel of interest. Must be 1 for SMUs with only one channel and 1 or 2 for SMUs with two channels. Default is 1. (default)

        Returns
        -------
        None
        """
        if self._sweep_mode == 0:  # 0 (VV mode)
            if not self._pseudo_bias_mode:  # 0 (current bias)
                logging.error(
                    '{!s}: Cannot set voltage value of channel {!s} in the current bias'.format(__name__, channel))
                raise SystemError(
                    '{!s}: Cannot set voltage value of channel {!s} in the current bias'.format(__name__, channel))
            elif self._pseudo_bias_mode:  # 1 (voltage bias)
                return self._SMU.ramp_bias(stop=stop * self._Vdiv, step=step * self._Vdiv, step_time=step_time,
                                           channel=channel)
        elif self._sweep_mode == 1:  # 1 (IV mode)
            logging.error(
                '{!s}: Cannot set voltage value of channel {!s} in the current bias'.format(__name__, channel))
            raise SystemError(
                '{!s}: Cannot set voltage value of channel {!s} in the current bias'.format(__name__, channel))
        elif self._sweep_mode == 2:  # 2 (VI-mode)
            return self._SMU.ramp_bias(stop=stop, step=step, step_time=step_time, channel=channel)

    def ramp_current(self, stop, step, step_time=0.1, channel=1):
        """
        Ramps current of channel <channel> from recent value to stop value <stop> with step size <step> and step time <step_time> according to bias_mode.
        This can only be set in the effective current bias mode, defined by <self.sweep_mode> and <self.pseudo_bias_mode>, and considers an external voltage-controlled current source <self._dAdV>.

        Parameters
        ----------
        stop: float
            Stop voltage value
        step: float
            Step voltage size.
        step_time: float
            Sleep time between voltage staircase ramp.
        channel: int
            Number of channel of interest. Must be 1 for SMUs with only one channel and 1 or 2 for SMUs with two channels. Default is 1. (default)

        Returns
        -------
        None
        """
        if self._sweep_mode == 0:  # 0 (VV mode)
            if not self._pseudo_bias_mode:  # 0 (current bias)
                return self._SMU.ramp_bias(stop=stop / self._dAdV, step=step / self._dAdV, step_time=step_time,
                                           channel=channel)
            elif self._pseudo_bias_mode:  # 1 (voltage bias)
                logging.error(
                    '{!s}: Cannot set current value of channel {!s} in the voltage bias'.format(__name__, channel))
                raise SystemError(
                    '{!s}: Cannot set current value of channel {!s} in the voltage bias'.format(__name__, channel))
        elif self._sweep_mode == 1:  # 1 (IV mode)
            return self._SMU.ramp_bias(stop=stop, step=step, step_time=step_time, channel=channel)
        elif self._sweep_mode == 2:  # 2 (VI-mode)
            logging.error(
                '{!s}: Cannot set current value of channel {!s} in the voltage bias'.format(__name__, channel))
            raise SystemError(
                '{!s}: Cannot set current value of channel {!s} in the voltage bias'.format(__name__, channel))

    def set_sweep_parameters(self, sweep):
        """
        Sets sweep parameters <sweep> and prepares instrument for the set sweep mode.

        Parameters
        ----------
        sweep: array_likes of floats
            Sweep range containing start, stop and step size (e.g. sweep object using qkit.measure.transport.transport.sweep class)

        Returns
        -------
        None
        """
        if not self._sweep_mode:  # 0 (VV-mode)
            if not self._pseudo_bias_mode:  # 0 (current bias)
                sweep = np.array(sweep).astype(float) / self._dAdV
                sweep[2] = np.abs(sweep[2])
                # alternative: self._sweep = np.array((lambda a: [a[0], a[1], np.abs(a[-1])])(sweep)).astype(float)/self._dAdV
            elif self._pseudo_bias_mode:  # 0 (voltage bias)
                sweep = np.array(sweep).astype(float) * self._Vdiv
                sweep[2] = np.abs(sweep[2])
        return self._SMU.set_sweep_parameters(sweep)

    def get_tracedata(self):
        """
        Starts bias sweep with already set parameters and gets trace data of bias <bias_values> and sense <sense_values> in the set sweep mode.

        Parameters
        ----------
        None

        Returns
        -------
        bias_values: numpy.array(float)
            Measured bias values.
        sense_values: numpy.array(float)
            Measured sense values.
        """
        bias_values, sense_values = self._SMU.get_tracedata()
        if self._sweep_mode == 0:  # IV-mode
            if not self._pseudo_bias_mode:  # 0 (current bias)
                I_values = bias_values * self._dAdV
                V_values = sense_values / self._amp
            elif self._pseudo_bias_mode:  # 1 (voltage bias)
                V_values = bias_values / self._Vdiv
                I_values = sense_values / self._dVdA
        elif self._sweep_mode == 1:  # IV-mode
            I_values = bias_values
            V_values = sense_values
        elif self._sweep_mode == 2:  # VI-mode
            V_values = bias_values
            I_values = sense_values
        return I_values, V_values

    def take_IV(self, sweep):
        """
        Takes IV curve with sweep parameters <sweep> in set sweep mode.

        Parameters
        ----------
        sweep: array_likes of floats
            Sweep range containing start, stop and step size (e.g. sweep object using qkit.measure.transport.transport.sweep class)

        Returns
        -------
        bias_values: numpy.array(float)
            Measured bias values.
        sense_values: numpy.array(float)
            Measured sense values.
        """
        self.set_sweep_parameters(sweep=sweep)
        return self.get_tracedata()

    def set_defaults(self, pseudo_bias_mode=None, SMU=True, sweep_mode=None):
        """
        Sets default settings for different pseudo bias modes <pseudo_bias_mode> and optional of the used source measure unit <SMU> of channel <channel>, too. Actual settings are:
        default parameters

        Parameters
        ----------
        pseudo_bias_mode: int
            Pseudo bias mode denoting current and voltage bias. Possible values are None (before used <self._sweep_mode>), 0 (current bias) and 1 (voltage bias). Default is None.
        SMU: bool
            Source measure unit. Default is True.
        sweep_mode: int
            Sweep mode denoting bias and sense modes. Possible values are None (before used <self._sweep_mode>), 0 (VV-mode), 1 (IV-mode) or 2 (VI-mode). Default is 1

        Returns
        -------
        None
        """
        # distinguish different pseudo bias modes
        if pseudo_bias_mode is not None:
            self.set_pseudo_bias_mode(pseudo_bias_mode)
        # set values
        for key_parameter, val_parameter in self._defaults[self._pseudo_bias_mode].items():
            eval('self.set_{:s}(val={!s})'.format(key_parameter, val_parameter))
        # set defaults of SMU according to sweep_mode
        if SMU:
            if sweep_mode is not None:
                self.set_sweep_mode(sweep_mode)
            self._SMU.set_defaults(sweep_mode=sweep_mode)

    def __set_defaults_docstring(self):
        '''
        Sets docstring of "set_defaults" method with actual default settings.

        Parameters
        ----------
        None

        Returns
        -------
        None
        '''
        new = 'pseudo bias mode: {:d}:\n'.format(self._pseudo_bias_mode) + ''.join(
            ['\t\t{:s}: {:1.2e}\n'.format(key_parameter, val_parameter) for key_parameter, val_parameter in
             self._defaults[self._pseudo_bias_mode].items()])
        self.set_defaults.__func__.__doc__ = self.set_defaults.__func__.__doc__.replace(self._default_str, new)
        self._default_str = new
        return

    def get_all(self, SMU=True, channel=1):
        """
        Prints all settings of channel <channel> and optional of the used source measure unit <SMU>, too.

        Parameters
        ----------
        SMU: bool
            Source measure unit. Default is True.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.

        Returns
        -------
        None
        """
        logging.info(__name__ + ': Get all')
        if self.get_BW() is not None:
            print('bandwidth          = {:1.0e}Hz'.format(self.get_BW()))
        print('dAdV               = {:1.0e}A/V'.format(self.get_dAdV()))
        print('amp factor         = {:1.0e}'.format(self.get_amp()))
        print('amp time constant  = {:1.0e}s'.format(self.get_tau_amp()))
        print('dVdA               = {:1.0e}V/A'.format(self.get_dVdA()))
        print('voltage divider    = {:1.0e}'.format(self.get_Vdiv()))
        print('sweep mode         = {:d} ({:s})'.format(self._sweep_mode, self._sweep_modes[self._sweep_mode]))
        print('pseudo bias mode   = {:d} ({:s})'.format(self._pseudo_bias_mode,
                                                        self._pseudo_bias_modes[self._pseudo_bias_mode]))
        if SMU:
            self._SMU.get_all(channel=channel)
        return

    def reset(self, SMU=True):
        """
        Resets internal variables for external tunnel electronic to default conditions and optional of the used source measure unit <SMU>, too.
        Resets the instrument or a single channel to .

        Parameters
        ----------
        SMU: bool
            Source measure unit. Default is True.

        Returns
        -------
        None
        """
        self.set_BW()
        self.set_dAdV()
        self.set_amp()
        self.set_tau_amp()
        self.set_dVdA()
        self.set_Vdiv()
        if SMU:
            self._SMU.reset()

    def get_parameters(self):
        """
        Gets a parameter list <parlist> of measurement specific setting parameters.
        Needed for .set-file in 'write_additional_files', if qt parameters are not used.

        Parameters
        ----------
        None

        Returns
        -------
        parlist: dict
            Parameter names as keys, corresponding channels of interest as values.
        """
        parlist = {# 'measurement_mode': self._sweep_channels,
                   # 'bias_mode': self._sweep_channels,
                   # 'sense_mode': self._sweep_channels,
                   # 'bias_range': self._sweep_channels,
                   # 'sense_range': self._sweep_channels,
                   # 'bias_delay': self._sweep_channels,
                   # 'sense_delay': self._sweep_channels,
                   # 'sense_average': self._sweep_channels,
                   # 'plc': [None],
                   # 'sense_nplc': self._sweep_channels,
                   'sweep_mode': [None],
                   'pseudo_bias_mode': [None],
                   'BW': [None]
                   }
        if not self._pseudo_bias_mode:  # 0 (current bias)
            parlist['dAdV'] = [None]
            parlist['amp'] = [None]
            # parlist['current'] = self._sweep_channels
        elif self._pseudo_bias_mode:  # 1 (voltage bias)
            parlist['dVdA'] = [None]
            parlist['Vdiv'] = [None]
            # parlist['voltage'] = self._sweep_channels
        return parlist

    def get(self, param, **kwargs):
        """
        Gets the current parameter <param> by evaluation 'get_'+<param> and corresponding channel if needed
        In combination with <self.get_parameters> above.

        Parameters
        ----------
        param: str
            Parameter of interest.
        channels: array_likes
            Number of channels of interest. Must be in {1, 2} for channel specific parameters or None for channel independant (global) parameters.

        Returns
        -------
        parlist: dict
            Parameter names as keys, values of corresponding channels as values.
        """
        channels = kwargs.get('channels')
        if channels != [None]:
            return tuple(eval('map(lambda channel, self=self: self.get_{:s}(channel=channel), [{:s}])'.format(param, ', '.join(map(str, channels)))))
            #return tuple([eval('self.get_{:s}(channel={!s})'.format(param, channel)) for channel in channels])
        else:
            return tuple(eval('map(lambda channel, self=self: self.get_{:s}(), [None])'.format(param)))
            #return eval('self.get_{:s}()'.format(param))



from lib_oled96 import ssd1306
from PIL import ImageFont
from smbus import SMBus
import logging
import numpy as np

i2cBus = SMBus(1)
oled = ssd1306(i2cBus)
draw = oled.canvas

fnt = ImageFont.truetype('FreeMonoBold.ttf', 17)

width = 16
height = 16
pos = [[], [4, 26], [24, 26], [44, 26], [4, 46], [24, 46], [44, 46]]


def draw_line():
    draw.rectangle((63, 0, 65, 64), outline=1, fill=1)


def draw_config(switch, config):
    add = 64 if switch == 1 else 0
    draw.text((5 + add, 5), "Sw: " + "AB"[switch], fill=1, font=fnt)
    for i in range(1, 7):
        fill = config[i - 1]
        if fill is None:
            draw.rectangle((pos[i][0] + add, pos[i][1], pos[i][0] + width + add, pos[i][1] + height), outline=1, fill=0)
            draw.text((pos[i][0] + 3 + add, pos[i][1] - 1), "?", fill=1, font=fnt)
        else:
            draw.rectangle((pos[i][0] + add, pos[i][1], pos[i][0] + width + add, pos[i][1] + height), outline=fill, fill=fill)
            draw.text((pos[i][0] + 3 + add, pos[i][1] - 1), str(i), fill=not fill, font=fnt)


import RPi.GPIO as GPIO
import time
from threading import Lock

# Set GPIO Mode
GPIO.setmode(GPIO.BOARD)

# GPIO Setup Pins
GPIO_OpAmp = 29  # OpAmp Signal
GPIO_SigInv = 31  # Invert polarity. Inverted when PIN is OFF!
PINS = [[32, 33, 36, 35, 38, 40], [7, 11, 12, 15, 16, 18]]
GPIO_ComGnd = 22  # Switch common Ground connection On/Off

# Set direction of GPIO Ports
for P in [GPIO_OpAmp, GPIO_SigInv, GPIO_ComGnd] + PINS[0] + PINS[1]:
    GPIO.setup(P, GPIO.OUT)


class Switch(object):
    def __init__(self):
        self._sw_pos = [[None] * 6, [None] * 6]
        self._pulse_len = [.08] * 2
        self._sleep_time = .2
        self._lock = Lock()
        self._update_display()
    
    def _pulse(self, switch, port, invert=False):
        with self._lock:
            if GPIO.input(GPIO_OpAmp):
                raise ValueError("OpAmp still enabled!")
            #GPIO.output(GPIO_ComGnd, True)
            GPIO.output(PINS[switch][port], True)
            GPIO.output(GPIO_SigInv, not invert)
            time.sleep(self._sleep_time)
            
            GPIO.output(GPIO_OpAmp, True)
            time.sleep(self._pulse_len[switch])
            GPIO.output(GPIO_OpAmp, False)
            
            time.sleep(self._sleep_time)
            GPIO.output(PINS[switch][port], False)
            GPIO.output(GPIO_SigInv, False)
            # Common ground is not switched off here.
            # GPIO.output(GPIO_ComGnd, False)
    
    def disable(self, switch, port):
        self._pulse(switch, port, True)
        self._sw_pos[switch][port] = False
        self._update_display()
    
    def enable(self, switch, port):
        self._pulse(switch, port, False)
        self._sw_pos[switch][port] = True
        #GPIO.output(GPIO_ComGnd, False)
        self._update_display()
    
    def switch_to(self, switch, port):
        if None in self._sw_pos[switch]:
            logging.warning("Not all switch positions are known for switch {0}. If you can tolerate heating, please use .reset_all(switch={0}).".format(switch))
        for i in np.where(self._sw_pos[switch])[0]:
            self.disable(switch, i)
        self.enable(switch, port)
    
    def reset_all(self, switch=None):
        if switch is None:
            self.reset_all(0)
            self.reset_all(1)
        else:
            for i in range(6):
                self.disable(switch, i)
            GPIO.output(GPIO_ComGnd, False)
        self._update_display()
    
    def get_switch_position(self, switch):
        return list(np.where(self._sw_pos[switch])[0])
    
    def set_pulse_length(self, length, switch=None):
        if switch is None:
            self._pulse_len = [length] * 2
        elif switch in [0, 1]:
            self._pulse_len[switch] = length
        else:
            raise ValueError("Only switch 0 or 1 are valid.")
    
    def get_pulse_length(self):
        return self._pulse_len
    
    def _update_display(self):
        oled.cls()
        draw_line()
        draw_config(0, self._sw_pos[0])
        draw_config(1, self._sw_pos[1])
        oled.display()
    
    def __del__(self):
        oled.cls()
        draw.text((20, 10), "SERVER", fill=1, font=fnt)
        draw.text((20, 25), "DOWN", fill=1, font=fnt)
        oled.display()



# NI_DAQ.py, National Instruments Data AcQuisition instrument driver
# Reinier Heeres <reinier@heeres.eu>, 2009
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import qkit
from qkit.core.instrument_base import Instrument
import types
from qkit.instruments import dll_support_nidaq as nidaq


def _get_channel(devchan):
    if not '/' in devchan:
        return devchan
    parts = devchan.split('/')
    if len(parts) != 2:
        return devchan
    return parts[1]


class NI_DAQ(Instrument):

    def __init__(self, name, id):
        Instrument.__init__(self, name, tags=['physical'])

        self._id = id

        for ch_in in self._get_input_channels():
            ch_in = _get_channel(ch_in)
            self.add_parameter(ch_in,
                               flags=Instrument.FLAG_GET,
                               type=float,
                               units='V',
                               tags=['measure'],
                               get_func=self.do_get_input,
                               channel=ch_in)

        for ch_out in self._get_output_channels():
            ch_out = _get_channel(ch_out)
            self.add_parameter(ch_out,
                               flags=Instrument.FLAG_SET,
                               type=float,
                               units='V',
                               tags=['sweep'],
                               set_func=self.do_set_output,
                               channel=ch_out)

        for ch_ctr in self._get_counter_channels():
            ch_ctr = _get_channel(ch_ctr)
            self.add_parameter(ch_ctr,
                               flags=Instrument.FLAG_GET,
                               type=int,
                               units='#',
                               tags=['measure'],
                               get_func=self.do_get_counter,
                               channel=ch_ctr)
            self.add_parameter(ch_ctr + "_src",
                               flags=Instrument.FLAG_SET | Instrument.FLAG_SOFTGET,
                               type=str,
                               set_func=self.do_set_counter_src,
                               channel=ch_ctr)

        self.add_parameter('chan_config',
                           flags=Instrument.FLAG_SET | Instrument.FLAG_SOFTGET,
                           type=str,
                           option_list=('Default', 'RSE', 'NRSE', 'Diff', 'PseudoDiff'))

        self.add_parameter('count_time',
                           flags=Instrument.FLAG_SET | Instrument.FLAG_SOFTGET,
                           type=float,
                           units='s')

        self.add_function('reset')
        self.add_function('digital_out')

        self.reset()
        self.set_chan_config('RSE')
        self.set_count_time(0.1)
        self.get_all()

    def get_all(self):
        ch_in = [_get_channel(ch) for ch in self._get_input_channels()]
        self.get(ch_in)

    def reset(self):
        '''Reset device.'''
        nidaq.reset_device(self._id)

    def _get_input_channels(self):
        return nidaq.get_physical_input_channels(self._id)

    def _get_output_channels(self):
        return nidaq.get_physical_output_channels(self._id)

    def _get_counter_channels(self):
        return nidaq.get_physical_counter_channels(self._id)

    def do_get_input(self, channel):
        devchan = '%s/%s' % (self._id, channel)
        return nidaq.read(devchan, config=self._chan_config)

    def do_set_output(self, val, channel):
        devchan = '%s/%s' % (self._id, channel)
        return nidaq.write(devchan, val)

    def do_set_chan_config(self, val):
        self._chan_config = val

    def do_set_count_time(self, val):
        self._count_time = val

    def do_get_counter(self, channel):
        devchan = '%s/%s' % (self._id, channel)
        src = self.get(channel + "_src")
        if src is not None and src != '':
            src = '/%s/%s' % (self._id, src)
        return nidaq.read_counter(devchan, src=src, freq=1 / self._count_time)

    def read_counters(self, channels):
        chans = []
        srcs = []
        for chan in channels:
            chans.append('%s/%s' % (self._id, chan))
            srcs.append(self.get(chan + "_src"))
        return nidaq.read_counters(chans, src=srcs, freq=1.0 / self._count_time)

    # Dummy
    def do_set_counter_src(self, val, channel):
        return True

    def digital_out(self, lines, val):
        devchan = '%s/%s' % (self._id, lines)
        return nidaq.write_dig_port8(devchan, val)


def detect_instruments():
    '''Refresh NI DAQ instrument list.'''

    for name in nidaq.get_device_names():
        qkit.instruments.create('NI%s' % name, 'NI_DAQ', id=name)



# AS @ KIT 04/2015
# modified JB 04/2016
# modified MP 05/2017 switch from pickle to JSON
# Sample Class to define all values necessary to measure a sample

import json, pickle
import logging
import os, copy

from qkit.measure.json_handler import QkitJSONEncoder, QkitJSONDecoder
from qkit.storage.hdf_DateTimeGenerator import DateTimeGenerator as dtg

import qkit

class Sample(object):
    '''
    Sample Class to define all values necessary to measure a sample
    '''
    def __init__(self, path=None):
        self.name = 'Arbitrary Sample'
        self.comment = ''
        if path:
            self.load(path)
            
    def get(self,argname,default=None):
        '''
        Gets an argument of the sample and returns default, if it does not exist.
        '''
        return self.__dict__.get(argname,default)
        

    def set_times(self,tpi):
        '''
        pass tpi to this function and it will update tpi as well as tpi2 = tpi/2
        '''
        self.tpi  = tpi
        self.tpi2 = tpi/2.

    def get_all(self):
        '''
        return all keys and entries of sample instance
        '''
        msg = ""
        copydict = copy.copy(self.__dict__)
        for key in sorted(copydict):
            msg+= str(key) + ":   " + str(copydict[key])+"\n"
        return msg

    def save(self,filename=None):
        '''
        save sample object.
        Filename can either be:
            - a full (absolute) path to the sample file or
            - any other string that is added to the filename in the data dir
            - None (default), the the sample object is stored in the data dir.
            
        returns the location of the sample file
        '''
        if filename is None or not(os.path.isabs(filename)):
            d = dtg()
            path = d.new_filename(filename)['_folder']
        else:
            path = filename

        if not os.path.exists(os.path.split(path)[0]):
            os.makedirs(os.path.split(path)[0])

        if not os.path.splitext(path)[-1] == '.sample':
            path = path + '.sample'

        copydict = copy.copy(self.__dict__)
        with open(path,'w+') as filehandler:
            json.dump(obj=copydict, fp=filehandler, cls=QkitJSONEncoder, indent = 4, sort_keys=True)
        return path

    def load(self, filename):
        '''
        load sample keys and entries to current sample instance
        '''

        if not os.path.isabs(filename):
            filename = os.path.join(qkit.cfg.get('datadir'),filename)

        try:
            with open(filename) as filehandle:
                self._load_legacy_pickle(filehandle)
            return
        except: pass

        with open(filename) as filehandle:
            self.__dict__ = json.load(filehandle, cls = QkitJSONDecoder)

    def _load_legacy_pickle(self, filehandle):
        self.__dict__ = pickle.loads(filehandle.read().split('<PICKLE PACKET BEGINS HERE>')[1].strip().replace('\r\n', '\n'))

        copydict = copy.copy(self.__dict__)
        for key in sorted(copydict):
            if ('xxxx'+str(copydict[key]))[-4:] == ' ins':   #instrument
                self.__dict__[key] = qkit.instruments.get(copydict[key][:-4])
